"""Gets GPU information"""
import re
import math
import logging
from shutil import which
from lnxlink.modules.scripts.helpers import import_install_package, syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "GPU"
        self.lnxlink = lnxlink
        self._requirements()
        self.gpu_ids = {"amd": 0}
        try:
            self.gpu_ids["amd"] = self.lib["amd"].detect_gpus()
        except Exception as err:
            logger.error("Can't get AMD GPU: %s", err)
        if which("nvidia-smi") is not None:
            try:
                self.gpu_ids["nvidia"] = len(list(self.lib["nvidia"].get_gpus()))
            except Exception as err:
                logger.error("Found nvidia-smi, but there was an error: %s", err)
                self.gpu_ids["nvidia"] = 0
        else:
            self.gpu_ids["nvidia"] = 0
        if len(self.gpu_ids["amd"]) == 0 and len(self.gpu_ids["nvidia"]) == 0:
            raise SystemError("No GPU found")

    def _requirements(self):
        self.lib = {
            "amd": import_install_package("pyamdgpuinfo", ">=2.1.4"),
            "nvidia": import_install_package("nvsmi", ">=0.4.2"),
        }

    def get_info(self):
        """Gather information from the system"""
        gpus = {}
        for gpu_id in range(self.gpu_ids["amd"]):
            amd_gpu = self.lib["amd"].get_gpu(gpu_id)
            gpus[f"amd_{gpu_id}"] = {
                "load": min(100, amd_gpu.query_load() * 100),
                "attributes": {
                    "Name": amd_gpu.name,
                    "VRAM usage": amd_gpu.query_vram_usage(),
                    "GTT usage": amd_gpu.query_gtt_usage(),
                    "Temperature": amd_gpu.query_temperature(),
                    "Power": amd_gpu.query_power(),
                    "Voltage": amd_gpu.query_graphics_voltage(),
                },
            }
        for gpu_id in range(self.gpu_ids["nvidia"]):
            nvidia_gpu = list(self.lib["nvidia"].get_gpus())[gpu_id]
            gpu_util = nvidia_gpu.gpu_util
            gpu_util = self._older_gpu_load(gpu_id, gpu_util)
            gpus[f"nvidia_{gpu_id}"] = {
                "load": gpu_util,
                "memory": round(nvidia_gpu.mem_util, 0),
                "temperature": nvidia_gpu.temperature,
                "attributes": {
                    "Name": nvidia_gpu.name,
                },
            }
        return gpus

    def _older_gpu_load(self, gpu_id, gpu_util):
        """For older GPUs, use nvidia-settings to get gpu usage"""
        if math.isnan(gpu_util):
            gpu_util = None
            if which("nvidia-settings") is not None:
                display = self.lnxlink.display
                if display:
                    settings_out, _, _ = syscommand(
                        f"nvidia-settings -q '[gpu:{gpu_id}]/GPUUtilization' --display {display}"
                    )
                    match = re.findall(r"graphics=(\d+)", settings_out)
                    if match:
                        gpu_util = int(match[0])
            else:
                logger.error(
                    "Older NVIDIA GPUs need nvidia-settings which is not installed."
                )
        return gpu_util

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for gpu_id in range(self.gpu_ids["amd"]):
            discovery_info[f"GPU AMD {gpu_id}"] = {
                "type": "sensor",
                "icon": "mdi:expansion-card-variant",
                "unit": "%",
                "state_class": "measurement",
                "value_template": f"{{{{ value_json.amd_{gpu_id}.load }}}}",
                "attributes_template": f"{{{{ value_json.amd_{gpu_id}.attributes | tojson }}}}",
                "enabled": True,
            }
        for gpu_id in range(self.gpu_ids["nvidia"]):
            for expose, unit in (("load", "%"), ("memory", "%"), ("temperature", "°C")):
                discovery_info[f"GPU NVIDIA {gpu_id} {expose}"] = {
                    "type": "sensor",
                    "icon": "mdi:expansion-card-variant",
                    "unit": unit,
                    "state_class": "measurement",
                    "value_template": f"{{{{ value_json.nvidia_{gpu_id}.{expose} }}}}",
                    "attributes_template": f"{{{{ value_json.nvidia_{gpu_id}.attributes | tojson }}}}",
                    "enabled": True,
                }
        return discovery_info



"""Checks if the Speaker is used"""
import glob
import json
import re
from lnxlink.modules.scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Speaker used"
        self.lnxlink = lnxlink

        _, _, returncode = syscommand(
            "which pactl && pactl -f json list short sink-inputs", ignore_errors=True
        )
        self.use_pactl = False
        if returncode == 0:
            self.use_pactl = True

    def get_info(self):
        """Gather information from the system"""
        if self.use_pactl:
            stdout, _, _ = syscommand(
                "pactl -f json list short sink-inputs",
            )
            # Replace 0,00 values with 0.00
            stdout = re.sub(r"(\s*[+-]?[0-9]+),([0-9]+\s*)", r"\1.\2", stdout)
            data = json.loads(stdout)
            if len(data) > 0:
                return "ON"
            return "OFF"
        speakers = glob.glob("/proc/asound/card*/pcm*/sub*/status", recursive=True)
        for speaker in speakers:
            with open(speaker, encoding="UTF-8") as speaker_content:
                speaker_status = speaker_content.read().strip().lower()
                if speaker_status != "closed":
                    return "ON"
        return "OFF"

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Speaker used": {
                "type": "binary_sensor",
                "icon": "mdi:speaker",
            },
        }



"""Auto load addons/modules"""
from importlib import import_module
from importlib.util import spec_from_file_location, module_from_spec
import time
import logging
import glob
import os
import requests

logger = logging.getLogger("lnxlink")


def autoload_modules(auto_exclude=None):
    """Gather a list of all modules"""
    if auto_exclude is None:
        auto_exclude = []
    modules = []
    modules_path = f"{os.path.dirname(__file__)}/*.py"
    for module_path in glob.glob(modules_path):
        module = os.path.basename(module_path)
        if "__" not in module and module.endswith(".py"):
            module = module.replace(".py", "")
            if module not in auto_exclude:
                modules.append(module)

    return modules


def parse_modules(list_modules=None, custom_modules=None, auto_exclude=None):
    """Import all modules and return them as a dict"""
    if list_modules is None:
        list_modules = autoload_modules(auto_exclude)
    if custom_modules is not None:
        list_modules.extend(custom_modules)
    modules = {}
    for module_name in list_modules:
        retries = 10
        while retries >= 0:
            try:
                if module_name.endswith(".py"):
                    if module_name.startswith("http"):
                        logger.info("Downloading custom module: %s", module_name)
                        module_data = requests.get(module_name, timeout=3).content
                        module_basename = os.path.basename(module_name)
                        module_name = f"/tmp/{module_basename}"
                        with open(module_name, "wb") as handler:
                            handler.write(module_data)
                    if os.path.isfile(module_name):
                        spec = spec_from_file_location("Addon", module_name)
                        module_spec = module_from_spec(spec)
                        spec.loader.exec_module(module_spec)
                        addon = getattr(module_spec, "Addon")
                        module_name = os.path.basename(module_name).split(".py")[0]
                    else:
                        logger.error("Can't find custom module: %s", module_name)
                        retries = -1
                        continue
                else:
                    addon = getattr(import_module(f"{__name__}.{module_name}"), "Addon")
                addon.service = module_name
                modules[module_name] = addon
                retries = -1
            except ModuleNotFoundError as err:
                logger.error(
                    "Addon %s is not supported, please remove it from your config: %s",
                    module_name,
                    err,
                )
                retries = -1
            except Exception as err:
                logger.error(
                    "Error with module %s: %s", module_name, err, exc_info=True
                )
                time.sleep(2)
                retries -= 1
    logger.debug("Found addons: %s", ", ".join(modules))
    return modules



"""Gets Interface names with their IP information"""
import logging
import psutil

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Interfaces"
        self.lnxlink = lnxlink
        self.interfaces = self._get_interfaces()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for device in self.interfaces:
            discovery_info[f"Interface {device}"] = {
                "type": "sensor",
                "icon": "mdi:lan-connect",
                "value_template": f"{{{{ value_json.get('{device}', {{}}).get('ipv4') }}}}",
                "attributes_template": f"{{{{ value_json.get('{device}', {{}}) | tojson }}}}",
                "enabled": True,
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        interfaces = self._get_interfaces()
        loaded = set(interfaces) - set(self.interfaces)
        unloaded = set(self.interfaces) - set(interfaces)
        for interface in unloaded:
            for addr in self.interfaces[interface]:
                self.interfaces[interface][addr] = None
            interfaces[interface] = self.interfaces[interface]
        self.interfaces = interfaces
        if len(loaded) > 0:
            self.lnxlink.setup_discovery()
        return self.interfaces

    def _bytetogb(self, byte):
        return round(byte / 1024 / 1024 / 1024, 0)

    def _get_interfaces(self):
        """Get a list of all interfaces"""
        interfaces = {}
        addrs = psutil.net_if_addrs()
        for interf, addr in addrs.items():
            if interf.startswith("veth"):
                continue
            for addr_item in addr:
                if addr_item.address not in ["127.0.0.1", "::1"]:
                    afinet = {
                        "AF_INET": "ipv4",
                        "AF_INET6": "ipv6",
                    }.get(addr_item.family.name)
                    if afinet is not None:
                        address = addr_item.address.replace(f"%{interf}", "")
                        interfaces.setdefault(interf, {})[afinet] = address
                        interfaces[interf][f"{afinet} NetMask"] = addr_item.netmask
        return interfaces



#!/usr/bin/env python3
"""Start the LNXlink service"""

import os
import sys
import time
import json
import threading
import logging
from logging.handlers import RotatingFileHandler
import argparse
import platform
import importlib.metadata
import ssl
import traceback

import yaml
import distro
import paho.mqtt.client as mqtt
from lnxlink import modules
from lnxlink import config_setup
from lnxlink.system_monitor import MonitorSuspend, GracefulKiller
from lnxlink.modules.scripts.helpers import syscommand


# Get the current version of the app
version = importlib.metadata.version(__package__ or __name__)
path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if os.path.exists(os.path.join(path, "lnxlink/edit.txt")):
    version += "+edit"
    git_hash, _, return_code = syscommand(f"git -C {path} rev-parse --short HEAD")
    if return_code == 0:
        version += f"-{git_hash}"

logger = logging.getLogger("lnxlink")


# pylint: disable=too-many-instance-attributes
class LNXlink:
    """Start LNXlink service that loads all modules and connects to MQTT"""

    version = version
    path = path

    def __init__(self, config_path):
        logger.info("LNXlink %s, Python %s", self.version, platform.python_version())
        self.config_path = config_path
        self.kill = None
        self.display = None
        self.inference_times = {}
        self.addons = {}
        self.prev_publish = {}
        self.update_change_interval = 900
        self.pref_topic = "lnxlink"
        if hasattr(mqtt, "CallbackAPIVersion"):
            self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)
        else:
            self.client = mqtt.Client()

        # Read configuration from yaml file
        self.config = self.read_config(config_path)

    def start(self, exclude_modules_arg):
        """Run each addon included in the modules folder"""
        conf_exclude = self.config["exclude"]
        conf_exclude = [] if conf_exclude is None else conf_exclude
        conf_exclude.extend(exclude_modules_arg)
        loaded_modules = modules.parse_modules(
            self.config["modules"], self.config["custom_modules"], conf_exclude
        )
        for _, addon in loaded_modules.items():
            try:
                tmp_addon = addon(self)
                self.addons[addon.service] = tmp_addon
            except Exception as err:
                logger.error(
                    "Error with addon %s, please remove it from your config: %s",
                    addon.service,
                    err,
                )
                logger.debug(
                    traceback.format_exc(),
                )
        logger.info("Loaded addons: %s", ", ".join(self.addons.keys()))

        # Setup MQTT
        return self.setup_mqtt()

    def publish_monitor_data(self, name, pub_data):
        """Publish info data to mqtt in the correct format"""
        subtopic = name.lower().replace(" ", "_")
        topic = f"{self.pref_topic}/monitor_controls/{subtopic}"
        if pub_data is None:
            return
        if isinstance(pub_data, bool):
            if pub_data is True:
                pub_data = "ON"
            if pub_data is False:
                pub_data = "OFF"
        if isinstance(pub_data, dict):
            if all(v is None for v in pub_data.values()):
                return
        if isinstance(pub_data, list):
            if all(v is None for v in pub_data):
                return
        if isinstance(pub_data, (dict, list)):
            pub_data = json.dumps(pub_data)

        # User option that checks and sends only the updated sensors
        update_change_time = time.time() - self.prev_publish.get("last_update", 0)
        if update_change_time > self.update_change_interval:
            self.prev_publish = {"last_update": time.time()}
        if (
            self.config["update_on_change"] or isinstance(pub_data, bytes)
        ) and self.prev_publish.get(topic) == pub_data:
            return

        self.prev_publish[topic] = pub_data
        self.client.publish(
            topic, payload=pub_data, retain=self.config["mqtt"]["lwt"]["retain"]
        )

    def run_module(self, name, method):
        """Runs the method of a module"""
        try:
            start_time = time.time()
            if isinstance(method, (dict, list, bool, bytes, int, str, float)):
                pub_data = method
            else:
                pub_data = method()
                diff_time = round(time.time() - start_time, 5)
                self.inference_times[name] = diff_time
            self.publish_monitor_data(name, pub_data)
        except Exception as err:
            logger.error(
                "Error with addon %s: %s, %s",
                name,
                err,
                traceback.format_exc(),
            )

    def monitor_run(self):
        """Gets information from each Addon and sends it to MQTT"""
        methods_to_run = []
        for _, addon in self.addons.items():
            if hasattr(addon, "get_info"):
                methods_to_run.append(
                    {
                        "name": addon.name,
                        "method": addon.get_info,
                    }
                )
        for method in methods_to_run:
            self.run_module(method["name"], method["method"])

    def monitor_run_thread(self):
        """Runs method to get sensor information every prespecified interval"""
        self.monitor_run()

        interval = self.config["update_interval"]
        if not self.kill:
            monitor = threading.Timer(interval, self.monitor_run_thread)
            monitor.start()

    def setup_mqtt(self):
        """Creates the mqtt object"""
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message

        keyfile = self.config["mqtt"]["auth"]["keyfile"]
        keyfile = None if keyfile == "" else keyfile
        certfile = self.config["mqtt"]["auth"]["certfile"]
        certfile = None if certfile == "" else certfile
        ca_certs = self.config["mqtt"]["auth"]["ca_certs"]
        ca_certs = None if ca_certs == "" else ca_certs
        use_cert = all(option is not None for option in [keyfile, certfile, ca_certs])
        use_tls = self.config["mqtt"]["auth"]["tls"]
        username = self.config["mqtt"]["auth"]["user"]
        password = self.config["mqtt"]["auth"]["pass"]
        use_userpass = all(option != "" for option in [username, password])

        if use_userpass:
            self.client.username_pw_set(username, password)
        if use_tls:
            cert_reqs = ssl.CERT_NONE
            if use_cert:
                cert_reqs = ssl.CERT_REQUIRED
            logger.info("Using MQTT ca_certs: %s", ca_certs)
            logger.info("Using MQTT certfile: %s", certfile)
            logger.info("Using MQTT keyfile: %s", keyfile)
            self.client.tls_set(
                ca_certs=ca_certs,
                certfile=certfile,
                keyfile=keyfile,
                cert_reqs=cert_reqs,
            )
            if ca_certs is None:
                self.client.tls_insecure_set(True)
        if self.config["mqtt"]["lwt"]["enabled"]:
            self.client.will_set(
                f"{self.pref_topic}/lwt",
                payload="OFF",
                qos=self.config["mqtt"]["lwt"]["qos"],
                retain=True,
            )
        try:
            self.client.connect(
                self.config["mqtt"]["server"], self.config["mqtt"]["port"], 60
            )
        except ssl.SSLCertVerificationError:
            logger.info("TLS not verified, using insecure connection instead")
            self.client.tls_insecure_set(True)
            self.client.connect(
                self.config["mqtt"]["server"], self.config["mqtt"]["port"], 60
            )
        except Exception as err:
            logger.error(
                "Error establishing connection to MQTT broker: %s, %s",
                err,
                traceback.format_exc(),
            )
            return False
        self.client.loop_start()
        return True

    def read_config(self, config_path):
        """Reads the config file and prepares module names for import"""
        with open(config_path, "r", encoding="utf8") as file:
            conf = yaml.load(file, Loader=yaml.FullLoader)

        self.pref_topic = f"{conf['mqtt']['prefix']}/{conf['mqtt']['clientId']}"
        self.pref_topic = self.pref_topic.lower()

        if conf["modules"] is not None:
            conf["modules"] = [x.lower().replace("-", "_") for x in conf["modules"]]
        return conf

    def on_connect(self, client, userdata, flags, rcode, *args):
        """Callback for MQTT connect which reports the connection status
        back to MQTT server"""
        logger.info("MQTT connection: %s", mqtt.connack_string(rcode))
        client.subscribe(f"{self.pref_topic}/commands/#")
        if self.config["mqtt"]["lwt"]["enabled"]:
            self.client.publish(
                f"{self.pref_topic}/lwt",
                payload="ON",
                qos=self.config["mqtt"]["lwt"]["qos"],
                retain=True,
            )
        if self.config["mqtt"]["discovery"]["enabled"]:
            self.setup_discovery()
        if self.kill is None:
            self.kill = False
            self.monitor_run_thread()

    def disconnect(self, *args):
        """Reports to MQTT server that the service has stopped"""
        logger.info("Disconnected from MQTT.")
        if self.config["mqtt"]["lwt"]["enabled"]:
            self.client.publish(
                f"{self.pref_topic}/lwt",
                payload="OFF",
                qos=self.config["mqtt"]["lwt"]["qos"],
                retain=True,
            )
        self.kill = True
        self.client.disconnect()

    def temp_connection_callback(self, status):
        """Report the connection status to MQTT server"""
        self.kill = True
        if self.config["mqtt"]["lwt"]["enabled"]:
            if status:
                logger.info("Power Down detected.")
                self.client.publish(
                    f"{self.pref_topic}/lwt",
                    payload="OFF",
                    qos=self.config["mqtt"]["lwt"]["qos"],
                    retain=True,
                )
            else:
                logger.info("Power Up detected.")
                if self.kill:
                    self.kill = False
                    self.monitor_run_thread()
                self.client.publish(
                    f"{self.pref_topic}/lwt",
                    payload="ON",
                    qos=self.config["mqtt"]["lwt"]["qos"],
                    retain=True,
                )

    def on_message(self, client, userdata, msg):
        """MQTT message is received with a module command to excecute"""
        topic = msg.topic.replace(f"{self.pref_topic}/commands/", "")
        message = msg.payload
        logger.info("Message received %s: %s", topic, message)
        try:
            message = json.loads(message)
        except json.decoder.JSONDecodeError:
            message = message.decode()
        except Exception as err:
            logger.debug("Error reading message: %s", err)

        select_service = topic.split("/")
        addon = self.addons.get(select_service[0])
        if addon is not None:
            if hasattr(addon, "start_control"):
                try:
                    result = addon.start_control(select_service, message)
                    if result is not None:
                        result_topic = (
                            f"{self.pref_topic}/command_result/{topic.strip('/')}"
                        )
                        self.client.publish(result_topic, payload=result, retain=False)
                except Exception as err:
                    logger.error(
                        "Couldn't run command for module %s: %s, %s",
                        addon,
                        err,
                        traceback.format_exc(),
                    )

    def setup_discovery_entities(self, addon, service, exp_name, options):
        """Send discovery information on Home Assistant for controls"""
        discovery_template = {
            "availability": {
                "topic": f"{self.pref_topic}/lwt",
                "payload_available": "ON",
                "payload_not_available": "OFF",
            },
            "device": {
                "identifiers": [self.config["mqtt"]["clientId"]],
                "name": self.config["mqtt"]["clientId"],
                "model": f"{distro.name()} {distro.version()}",
                "manufacturer": "LNXlink",
                "sw_version": version,
            },
        }
        control_name_topic = exp_name.lower().replace(" ", "_")
        subtopic = addon.name.lower().replace(" ", "_")
        if "method" in options or options.get("subtopic", False):
            subcontrol = exp_name.lower().replace(" ", "_")
            subtopic = f"{subtopic}/{subcontrol}"
        state_topic = f"{self.pref_topic}/monitor_controls/{subtopic}"
        command_topic = f"{self.pref_topic}/commands/{service}/{control_name_topic}/"

        lookup_options = {
            "value_template": {
                "value_template": options.get("value_template", ""),
            },
            "attributes_template": {
                "json_attributes_topic": state_topic,
                "json_attributes_template": options.get(
                    "attributes_template", "{{ value_json | tojson }}"
                ),
            },
            "icon": {"icon": options.get("icon", "")},
            "unit": {"unit_of_measurement": options.get("unit", "")},
            "title": {"title": options.get("title", "")},
            "entity_picture": {"entity_picture": options.get("entity_picture", "")},
            "device_class": {"device_class": options.get("device_class", "")},
            "state_class": {"state_class": options.get("state_class", "")},
            "entity_category": {"entity_category": options.get("entity_category", "")},
            "enabled": {"enabled_by_default": options.get("enabled", True)},
            "expire_after": {"expire_after": options.get("expire_after", "")},
            "install": {
                "command_topic": command_topic,
                "payload_install": options.get("install", ""),
            },
        }
        lookup_entities = {
            "sensor": {
                "state_topic": state_topic,
            },
            "binary_sensor": {
                "state_topic": state_topic,
            },
            "camera": {
                "topic": state_topic,
                "image_encoding": options.get("encoding"),
            },
            "image": {
                "image_topic": state_topic,
                "image_encoding": options.get("encoding"),
            },
            "update": {"state_topic": state_topic},
            "button": {
                "command_topic": command_topic,
                "payload_press": options.get("payload_press", "PRESS"),
            },
            "switch": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "payload_off": options.get("command_off", "OFF"),
                "payload_on": options.get("command_on", "ON"),
                "state_off": "OFF",
                "state_on": "ON",
            },
            "text": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "min": options.get("min", 0),
                "max": options.get("max", 255),
            },
            "number": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "min": options.get("min", 1),
                "max": options.get("max", 100),
                "step": options.get("step", 1),
            },
            "select": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "options": options.get("options", []),
            },
        }
        discovery = discovery_template.copy()
        discovery["name"] = exp_name
        discovery[
            "unique_id"
        ] = f"{self.config['mqtt']['clientId']}_{control_name_topic}"
        discovery.update(lookup_entities.get(options["type"], {}))
        for option in options:
            discovery.update(lookup_options.get(option, {}))

        if options["type"] not in lookup_entities:
            logger.error("Not supported: %s", options["type"])
            return
        if "value_template" in discovery and options["type"] in ["camera", "image"]:
            del discovery["json_attributes_topic"]
            del discovery["json_attributes_template"]
        self.client.publish(
            f"homeassistant/{options['type']}/lnxlink/{discovery['unique_id']}/config",
            payload=json.dumps(discovery),
            retain=True,
        )

    def setup_discovery(self):
        """First time setup of discovery for Home Assistant"""
        for service, addon in self.addons.items():
            if hasattr(addon, "exposed_controls"):
                for exp_name, options in addon.exposed_controls().items():
                    try:
                        self.setup_discovery_entities(addon, service, exp_name, options)
                    except Exception as err:
                        logger.error(
                            "%s: %s, %s", exp_name, err, traceback.format_exc()
                        )

    def restart_script(self):
        """Restarts itself"""
        logger.info("Restarting LNXlink...")
        os.execv(sys.executable, ["python"] + sys.argv)


def setup_logger(config_path, log_level):
    """Save logs on the same directory as the config file"""
    config_dir = os.path.dirname(os.path.realpath(config_path))
    start_sec = str(int(time.time()))[-4:]
    log_formatter = logging.Formatter(
        "%(asctime)s ["
        + start_sec
        + ":%(threadName)s.%(module)s.%(funcName)s.%(lineno)d] [%(levelname)s]  %(message)s"
    )

    file_handler = RotatingFileHandler(
        f"{config_dir}/lnxlink.log",
        maxBytes=5 * 1024 * 1024,
        backupCount=1,
    )
    logging.basicConfig(level=log_level)
    file_handler.setFormatter(log_formatter)
    logger.addHandler(file_handler)


def main():
    """Starts the app with some arguments"""
    description = (
        f"LNXlink {version} bridges this OS with Home Assistant through an MQTT broker."
    )
    parser = argparse.ArgumentParser(prog="lnxlink", description=description)
    parser.add_argument(
        "-c",
        "--config",
        help="Configuration file",
    )
    parser.add_argument(
        "-i",
        "--ignore-systemd",
        help="Runs without setting up SystemD service",
        action="store_true",
    )
    parser.add_argument(
        "-e",
        "--exclude",
        help="Exclude modules from running",
        default=[],
        type=lambda t: [s.strip() for s in t.split(",")],
    )
    parser.add_argument(
        "-l",
        "--logging",
        help="Set the logging level",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    parser.add_argument(
        "-s",
        "--setup",
        help="Runs only the setup configuration workflow",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--version",
        help="Shows current version of LNXlink",
        action="store_true",
    )
    args = parser.parse_args()

    if args.version:
        print(version)
        sys.exit()
    if args.config is None:
        parser.print_help()
        parser.exit("\nSomething went wrong, --config condition was not set")
    config_file = os.path.abspath(args.config)
    setup_logger(config_file, args.logging)
    config_setup.setup_config(config_file)
    if args.setup:
        logger.info("The configuration exists under the file: %s", config_file)
        sys.exit()
    if not args.ignore_systemd:
        config_setup.setup_systemd(config_file)
    else:
        logger.info(
            "By not setting up the SystemD, LNXlink won't be able to start on boot..."
        )

    lnxlink = LNXlink(config_file)

    # Monitor for system changes (Shutdown/Suspend/Sleep)
    monitor_suspend = MonitorSuspend(lnxlink.temp_connection_callback)
    monitor_suspend.start()
    monitor_gracefulkiller = GracefulKiller(lnxlink.temp_connection_callback)

    # Starts the main app
    start_status = lnxlink.start(args.exclude)
    if start_status:
        while not monitor_gracefulkiller.kill_now:
            time.sleep(1.0)
        monitor_suspend.stop()
        lnxlink.disconnect()
    else:
        monitor_suspend.stop()


if __name__ == "__main__":
    main()



"""Information and control of pulseaudio devices"""
import logging
from lnxlink.modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Audio Select"
        self.lnxlink = lnxlink
        self.devices = {
            "speaker": {},
            "microphone": {},
            "defaults": {"microphone": "", "speaker": ""},
            "changed": False,
        }
        self._requirements()
        self._get_devices()

    def _requirements(self):
        self.lib = {
            "pulsectl": import_install_package("pulsectl", ">=23.5.2"),
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {
            "Microphone Select": {
                "type": "select",
                "icon": "mdi:microphone-variant",
                "options": list(self.devices["microphone"].keys()),
                "value_template": "{{ value_json.microphone }}",
            },
            "Speaker Select": {
                "type": "select",
                "icon": "mdi:audio-input-rca",
                "options": list(self.devices["speaker"].keys()),
                "value_template": "{{ value_json.speaker }}",
            },
        }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        self._get_devices()
        if self.devices["changed"]:
            self.lnxlink.setup_discovery()
        return self.devices["defaults"]

    def start_control(self, topic, data):
        """Control system"""
        if topic[1] == "microphone_select":
            with self.lib["pulsectl"].Pulse("lnxlink_ctl") as pulse:
                pulse.default_set(
                    pulse.get_source_by_name(self.devices["microphone"][data])
                )
        if topic[1] == "speaker_select":
            with self.lib["pulsectl"].Pulse("lnxlink_ctl") as pulse:
                pulse.default_set(pulse.get_sink_by_name(self.devices["speaker"][data]))

    def _get_devices(self):
        """Get a list of all audio devices"""
        devices = {
            "speaker": {},
            "microphone": {},
            "defaults": {"microphone": "", "speaker": ""},
            "changed": False,
        }

        try:
            with self.lib["pulsectl"].Pulse("lnxlink") as pulse:
                for sink in pulse.sink_list():
                    devices["speaker"][sink.description] = sink.name
                for source in pulse.source_list():
                    devices["microphone"][source.description] = source.name

                server_info = pulse.server_info()
                for description, name in devices["speaker"].items():
                    if name == server_info.default_sink_name:
                        devices["defaults"]["speaker"] = description
                for description, name in devices["microphone"].items():
                    if name == server_info.default_source_name:
                        devices["defaults"]["microphone"] = description
        except Exception as err:
            logger.error("Pulseaudio is not installed on your system: %s", err)

        if len(devices["defaults"]["speaker"]) != len(
            self.devices["defaults"].get("speaker", [])
        ):
            devices["changed"] = True
        if len(devices["defaults"]["microphone"]) != len(
            self.devices["defaults"].get("microphone", [])
        ):
            devices["changed"] = True
        self.devices = devices

        return devices



"""Open URLs or files"""
import logging
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "xdg_open"
        if which("xdg-open") is None:
            raise SystemError("System command 'xdg-open' not found")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "XDG Open": {
                "type": "text",
                "icon": "mdi:file-find-outline",
            }
        }

    def start_control(self, topic, data):
        """Control system"""
        logger.info("xdg-open %s", data)
        syscommand(f"xdg-open {data}", background=True)



"""Uses xdotool to control mouse"""
import os
import time
import logging
import threading
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Mouse"
        self.lnxlink = lnxlink
        self.movement = [0, 0]
        if which("xdotool") is None:
            raise SystemError("System command 'xdotool' not found")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Mouse Coordinates": {
                "type": "text",
                "icon": "mdi:mouse-variant",
            },
            "Mouse Left": {
                "type": "button",
                "icon": "mdi:gamepad-circle-left",
            },
            "Mouse Right": {
                "type": "button",
                "icon": "mdi:gamepad-circle-right",
            },
            "Mouse Up": {
                "type": "button",
                "icon": "mdi:gamepad-circle-up",
            },
            "Mouse Down": {
                "type": "button",
                "icon": "mdi:gamepad-circle-down",
            },
            "Mouse Click": {
                "type": "button",
                "icon": "mdi:mouse-left-click",
            },
            "Mouse Click Right": {
                "type": "button",
                "icon": "mdi:mouse-right-click",
            },
        }

    def start_control(self, topic, data):
        """Control system"""
        if os.environ.get("DISPLAY") is None:
            if self.lnxlink.display is not None:
                os.environ["DISPLAY"] = self.lnxlink.display
                logger.info("Initializing empty DISPLAY environment variable")

        if topic[1] == "mouse_coordinates":
            if "," in data:
                coords = data.replace(" ", "").split(",")
            elif " " in data:
                coords = data.split(" ")
            syscommand(f"xdotool mousemove {coords[0]} {coords[1]}")
        elif topic[1] == "mouse_left":
            self._move([-1, 0])
        elif topic[1] == "mouse_right":
            self._move([1, 0])
        elif topic[1] == "mouse_up":
            self._move([0, -1])
        elif topic[1] == "mouse_down":
            self._move([0, 1])
        elif topic[1] == "mouse_click":
            syscommand("xdotool click 1")
        elif topic[1] == "mouse_click_right":
            syscommand("xdotool click 3")

    def _move(self, movement):
        if self.movement == movement:
            # Stops mouse if it is moving
            self.movement = [0, 0]
        else:
            self.movement = movement
            move_mouse_thr = threading.Thread(target=self._move_mouse)
            move_mouse_thr.start()

    def _move_mouse(self):
        movement = self.movement
        for i in range(1, 90):
            if self.movement != movement:
                # Stops mouse if a new move command is sent
                return
            move_x = movement[0] * i
            move_y = movement[1] * i
            syscommand(f"xdotool mousemove_relative -- {move_x} {move_y}")
            time.sleep(0.05)
        # Used if the user wants to run the same movement
        self.movement = [0, 0]



"""Monitors for shutdown/sleep events"""

import threading
import signal
import logging
import traceback

from lnxlink.modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class MonitorSuspend:
    """Monitor DBUS for Suspend and Shutdown events"""

    def __init__(self, callback):
        self.use = False
        self._requirements()
        if self.lib["dbus"] is None:
            logger.error("Can't use DBus for system monitor")
            return
        try:
            bus = self.lib["dbus"].connection.SystemMessageBus()

            proxy = bus.get_proxy(
                service_name="org.freedesktop.login1",
                object_path="/org/freedesktop/login1",
                interface_name="org.freedesktop.login1.Manager",
            )
            proxy.PrepareForSleep.connect(callback)
            proxy.PrepareForShutdown.connect(callback)

            self.loop = self.lib["dbus-loop"].loop.EventLoop()
            self.timer1 = threading.Thread(target=self.loop.run, daemon=True)
            self.use = True
        except Exception as err:
            logger.error(
                "Can't use DBus: %s, %s",
                err,
                traceback.format_exc(),
            )

    def _requirements(self):
        dbus = import_install_package("dasbus", ">=1.7", "dasbus.connection")
        if dbus is None:
            self.lib = {"dbus": dbus}
            return
        self.lib = {
            "dbus": dbus,
            "dbus-loop": import_install_package("dasbus", ">=1.7", "dasbus.loop"),
        }

    def start(self):
        """Start the timer of the thread"""
        if self.use:
            self.timer1.start()

    def stop(self):
        """Stop the timer of the thread"""
        if self.use:
            self.loop.quit()
            self.timer1.join()


class GracefulKiller:
    """Monitor Signal Termination"""

    def __init__(self, callback):
        """Checks for termination signals"""
        self.kill_now = False
        self.callback = callback
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, signum, frame):
        """Stop the app gracefully"""
        logger.info("stopped gracefully")
        self.callback(1)
        self.kill_now = True



"""Shutdown the system"""
import logging
from shutil import which
from lnxlink.modules.scripts.helpers import import_install_package, syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Shutdown"
        self.lnxlink = lnxlink
        if which("systemctl") is None and which("shutdown") is None:
            self.dbus = import_install_package("dasbus", ">=1.7", "dasbus.connection")

    def start_control(self, topic, data):
        """Control system"""
        self.lnxlink.temp_connection_callback(True)
        if which("systemctl") is not None:
            syscommand("systemctl poweroff")
        elif which("shutdown") is not None:
            syscommand("shutdown now")
        else:
            bus = self.dbus.connection.SystemMessageBus()
            proxy = bus.get_proxy(
                service_name="org.freedesktop.login1",
                object_path="/org/freedesktop/login1",
            )
            proxy.PowerOff(True)

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Shutdown": {
                "type": "button",
                "icon": "mdi:power",
            }
        }



"""Gets the active window"""
from lnxlink.modules.scripts.helpers import import_install_package


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Active Window"
        self.lnxlink = lnxlink
        self._requirements()

    def _requirements(self):
        self.lib = {
            "ewmh": import_install_package("ewmh", ">=0.1.6"),
            "xlib": import_install_package("python-xlib", ">=0.33", "Xlib.display"),
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Active Window": {
                "type": "sensor",
                "icon": "mdi:book-open-page-variant",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        if self.lnxlink.display is None:
            return ""
        display = self.lib["xlib"].display.Display(self.lnxlink.display)
        ewmh = self.lib["ewmh"].EWMH(_display=display)
        win = ewmh.getActiveWindow()
        window_name = ewmh.getWmName(win)
        if window_name is None:
            return None
        return window_name.decode()



"""Gets the idle time of the system"""
import logging
from lnxlink.modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")
logging.getLogger("dasbus").setLevel(logging.WARNING)


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Idle"
        self._requirements()
        if self.lib["dbus_idle"] is None:
            raise SystemError("Python package 'dbus_idle' can't be installed")

    def _requirements(self):
        self.lib = {
            "dbus_idle": import_install_package("dbus-idle", ">=2024.7.1", "dbus_idle"),
        }

    def get_info(self):
        """Gather information from the system"""
        idle_ms = self.lib["dbus_idle"].IdleMonitor().get_dbus_idle()
        if idle_ms is None:
            return None
        idle_sec = round(idle_ms / 1000, 0)
        return idle_sec

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Idle": {
                "type": "sensor",
                "icon": "mdi:timer-sand",
                "unit": "s",
                "state_class": "total_increasing",
                "device_class": "duration",
            },
        }



"""Turns on or off the screen"""
import re
import logging
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Screen OnOff"
        self.lnxlink = lnxlink
        if which("xset") is None:
            raise SystemError("System command 'xset' not found")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Screen OnOff": {
                "type": "switch",
                "icon": "mdi:monitor",
            }
        }

    def get_info(self):
        """Gather information from the system"""
        status = "ON"
        if self.lnxlink.display is not None:
            command = f"xset -display {self.lnxlink.display} q"
            stdout, _, _ = syscommand(command)
            match = re.findall(r"Monitor is (\w*)", stdout)
            if match:
                status = match[0].upper()
            else:
                logger.debug("Screen_onoff error: %s\n%s", command, stdout)
        else:
            logger.debug("No display variable found")

        return status

    def start_control(self, topic, data):
        """Control system"""
        if data.lower() == "off":
            syscommand(f"xset -display {self.lnxlink.display} dpms force off")
        elif data.lower() == "on":
            syscommand(f"xset -display {self.lnxlink.display} dpms force on")



"""Gets mount usage from all mounted directories"""
import logging
from lnxlink.modules.scripts.helpers import syscommand, import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Mounts"
        self.lnxlink = lnxlink
        self.lib = {}
        self.mounts = self._get_mounts()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for device in self.mounts:
            att_temp = f"{{{{ value_json.get('{device}', {{}}).get('attributes', {{}}) | tojson }}}}"
            discovery_info[f"Mount {device}"] = {
                "type": "sensor",
                "icon": "mdi:harddisk",
                "unit": "%",
                "entity_category": "diagnostic",
                "state_class": "measurement",
                "value_template": f"{{{{ value_json.get('{device}', {{}}).get('percent') }}}}",
                "attributes_template": att_temp,
                "enabled": True,
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        mounts = self._get_mounts()
        mounted = set(mounts) - set(self.mounts)
        unmounted = set(self.mounts) - set(mounts)
        for mount_name in unmounted:
            mounts[mount_name] = self.mounts[mount_name]
            self.mounts[mount_name]["attributes"]["connected"] = False
            self.mounts[mount_name]["percent"] = None
        self.mounts = mounts
        if len(mounted) > 0:
            self.lnxlink.setup_discovery()
        return self.mounts

    def _get_mounts(self):
        """Get a list of all mounts"""
        mounts = {}
        mount_directories = (
            self.lnxlink.config["settings"].get("mounts", {}).get("directories", [])
        )
        for mount_directory in mount_directories:
            unique_name, mount_data = self._calculate_size(mount_directory)
            mounts[unique_name] = mount_data

        if self.lnxlink.config["settings"].get("mounts", {}).get("autocheck", False):
            mounts.update(self._get_remote_mounts())

        return mounts

    def _calculate_size(self, mountpoint):
        unique_name = mountpoint
        unique_name = unique_name.replace(" ", "_")
        unique_name = unique_name.replace("=", "_")
        unique_name = unique_name.replace(".", "_")
        unique_name = unique_name.replace(",", "")
        unique_name = unique_name.replace("/", "")
        unique_name = unique_name.replace(",", "")
        unique_name = unique_name.replace(":", "")
        unique_name = unique_name.replace("-", "")

        data = {
            "percent": None,
            "attributes": {
                "mountpoint": mountpoint,
                "total": None,
                "filesystem": None,
                "mount_on": None,
                "connected": False,
            },
        }
        stdout, _, _ = syscommand(["df", mountpoint])
        if len(stdout.split("\n")) == 2:
            filesystem, size_str, _, _, used, mounton = stdout.split("\n")[1].split()
            data["percent"] = int(used.replace("%", ""))
            data["attributes"]["total"] = round(int(size_str) / 1024 / 1024, 0)
            data["attributes"]["filesystem"] = filesystem
            data["attributes"]["mount_on"] = mounton
            data["attributes"]["connected"] = True
        return unique_name, data

    def _requirements(self):
        self.lib["gi"] = import_install_package(
            "PyGObject", ">=3.44.0", "gi.repository.Gio"
        )

    def _get_remote_mounts(self):
        """Uses Gnome GVFS to find remote mounts mounted by the system"""
        if "gi" not in self.lib:
            self._requirements()

        mounts_dict = {}
        volume_monitor = self.lib["gi"].repository.Gio.VolumeMonitor.get()
        mounts = volume_monitor.get_mounts()
        for mount in mounts:
            name = mount.get_name()
            mounttype = mount.get_root().get_uri_scheme()
            native = mount.get_root().is_native()
            mountpoint = mount.get_root().get_path()
            if not native:
                unique_name, mount_data = self._calculate_size(mountpoint)
                if mount_data["percent"] is not None:
                    mounts_dict[unique_name] = mount_data
                    mounts_dict[unique_name]["attributes"]["name"] = name
                    mounts_dict[unique_name]["attributes"]["mount_type"] = mounttype
        return mounts_dict



"""Suspend/sleep the system"""
from lnxlink.modules.scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Suspend"

    def start_control(self, topic, data):
        """Control system"""
        syscommand("systemctl suspend")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Suspend": {
                "type": "button",
                "icon": "mdi:progress-clock",
            }
        }



"""Selects the OS to boot from on GRUB at the next restart"""
import os
import re
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink=None):
        """Setup addon"""
        self.name = "Boot Select"
        self.lnxlink = lnxlink
        if which("grub-editenv") is None:
            raise SystemError("System command 'grub-editenv' not found")
        self.options = self._get_grub_entries()

    def get_info(self):
        """Gather information from the system"""
        stdout, _, _ = syscommand("grub-editenv list")
        nextentry_pattern = re.compile(r"^next_entry=(\d+)")
        nextentry_match = re.match(nextentry_pattern, stdout)
        entry_ind = 0
        if nextentry_match:
            entry_ind = int(nextentry_match.group(1))

        if len(self.options) > 0:
            return self.options[entry_ind]
        return ""

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Boot Select": {
                "type": "select",
                "icon": "mdi:harddisk-plus",
                "options": self.options,
            }
        }

    def start_control(self, topic, data):
        """Control system"""
        ind = self.options.index(data)
        syscommand(f"sudo grub-reboot {ind}")

    def _get_grub_entries(self):
        """Get the grub entities in the correct order"""
        menu_pattern = re.compile("^menuentry '([^']*)'")
        submenu_pattern = re.compile("^submenu '([^']*)'")

        grub_entries = []

        if os.path.exists("/boot/grub/grub.cfg"):
            with open("/boot/grub/grub.cfg", encoding="UTF-8") as file:
                for line in file.readlines():
                    menu_entry_match = re.match(menu_pattern, line)
                    if menu_entry_match:
                        grub_entry = menu_entry_match.group(1)
                        grub_entries.append(grub_entry)

                    submenu_entry_match = re.match(submenu_pattern, line)
                    if submenu_entry_match:
                        grub_entry = submenu_entry_match.group(1)
                        grub_entries.append(grub_entry)

        return grub_entries



"""Gets disk usage from all disks"""
import logging

import psutil

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Disk Usage"
        self.lnxlink = lnxlink
        self.disks = self._get_disks()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for device in self.disks:
            att_temp = f"{{{{ value_json.get('{device}', {{}}).get('attributes', {{}}) | tojson }}}}"
            discovery_info[f"Disk {device}"] = {
                "type": "sensor",
                "icon": "mdi:harddisk",
                "unit": "%",
                "entity_category": "diagnostic",
                "state_class": "measurement",
                "value_template": f"{{{{ value_json.get('{device}', {{}}).get('percent') }}}}",
                "attributes_template": att_temp,
                "enabled": True,
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        disks = self._get_disks()
        mounted = set(disks) - set(self.disks)
        unmounted = set(self.disks) - set(disks)
        for disk_name in unmounted:
            disks[disk_name] = self.disks[disk_name]
            self.disks[disk_name]["attributes"]["connected"] = False
            self.disks[disk_name]["percent"] = None
        self.disks = disks
        if len(mounted) > 0:
            self.lnxlink.setup_discovery()
        return self.disks

    def _bytetogb(self, byte):
        return round(byte / 1024 / 1024 / 1024, 0)

    def _get_disks(self):
        """Get a list of all disks"""
        disks = {}
        disk_includes = (
            self.lnxlink.config["settings"]
            .get("disk_usage", [])
            .get("include_disks", [])
        )
        disk_excludes = (
            self.lnxlink.config["settings"]
            .get("disk_usage", [])
            .get("exclude_disks", [])
        )
        for disk in psutil.disk_partitions():
            matches = [
                "/snap/",
                "/docker/overlay",
                "/docker/btrfs",
            ]
            if any(x in disk.mountpoint for x in matches):
                continue
            if disk.fstype == "squashfs":
                continue
            if len(disk_includes) != 0:
                if not any(disk.device.startswith(x) for x in disk_includes):
                    continue
            if len(disk_excludes) != 0:
                if any(disk.device.startswith(x) for x in disk_excludes):
                    continue

            try:
                disk_stats = psutil.disk_usage(disk.mountpoint)
                device = disk.device.replace("/", "_").strip("_")
                disks[device] = {}
                disks[device]["percent"] = disk_stats.percent
                disks[device]["attributes"] = {}
                disks[device]["attributes"]["total"] = self._bytetogb(disk_stats.total)
                disks[device]["attributes"]["connected"] = True
                disks[device]["attributes"]["mountpoint"] = disk.mountpoint
            except Exception as err:
                logger.error(
                    "Error with disk usage [fstype: %s, device: %s, mountpoint: %s]: %s",
                    disk.fstype,
                    disk.device,
                    disk.mountpoint,
                    err,
                )
        return disks



"""Shows an image from the webcamera"""
import base64
from threading import Thread
from lnxlink.modules.scripts.helpers import import_install_package


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.lnxlink = lnxlink
        self.name = "Webcam"
        self.vid = None
        self._requirements()
        self.read_thr = None

    def _requirements(self):
        self.lib = {
            "cv2": import_install_package("opencv-python", ">=4.7.0.68", "cv2"),
        }

    def get_camera_frame(self):
        """Convert camera feed to Base64 text"""
        if self.vid is not None:
            while True:
                _, frame = self.vid.read()
                _, buffer = self.lib["cv2"].imencode(".jpg", frame)
                frame = base64.b64encode(buffer)
                self.lnxlink.run_module(f"{self.name}/Webcam feed", frame)

    def get_info(self):
        """Gather information from the system"""
        if self.vid is not None:
            return True
        return False

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Webcam": {
                "type": "switch",
                "icon": "mdi:webcam",
                "entity_category": "config",
            },
            "Webcam feed": {
                "type": "camera",
                "encoding": "b64",
                "subtopic": True,
            },
        }

    def start_control(self, topic, data):
        """Control system"""
        if data.lower() == "off":
            self.vid.release()
            self.vid = None
            if self.read_thr is not None:
                self.read_thr.join()
                self.read_thr = None
        elif data.lower() == "on":
            self.vid = self.lib["cv2"].VideoCapture(0)
            if self.read_thr is None:
                self.read_thr = Thread(target=self.get_camera_frame, daemon=True)
                self.read_thr.start()



"""Selects the Power Profile"""
import re
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink=None):
        """Setup addon"""
        self.name = "Power Profile"
        self.lnxlink = lnxlink
        if which("powerprofilesctl") is None:
            raise SystemError("System command 'powerprofilesctl' not found")
        self.options = self._get_power_profiles()

    def get_info(self):
        """Gather information from the system"""
        stdout, _, _ = syscommand("powerprofilesctl get")
        return stdout

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        if len(self.options) > 0:
            discovery_info = {
                "Power Profile": {
                    "type": "select",
                    "icon": "mdi:leaf",
                    "options": self.options,
                }
            }
        return discovery_info

    def start_control(self, topic, data):
        """Control system"""
        syscommand(f"powerprofilesctl set {data}")

    def _get_power_profiles(self):
        """Get the power profiles in the correct order"""
        profiles_pattern = re.compile(r"([\w-]+):\n")

        stdout, _, _ = syscommand("powerprofilesctl list")
        profiles = re.findall(profiles_pattern, stdout)

        return profiles



"""Gets the battery information of connected devices"""
from xml.etree import ElementTree
from lnxlink.modules.scripts.helpers import import_install_package


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Battery"
        self.lnxlink = lnxlink
        self._requirements()
        self.devices = self._get_devices()

    def _requirements(self):
        dasbus = import_install_package("dasbus", ">=1.7", "dasbus.connection")
        self.bus = dasbus.connection.SystemMessageBus()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for device in self.devices:
            att_temp = f"{{{{ value_json.get('{device}', {{}}).get('attributes', {{}}) | tojson }}}}"
            discovery_info[f"Battery {device}"] = {
                "type": "sensor",
                "unit": "%",
                "device_class": "battery",
                "value_template": f"{{{{ value_json.get('{device}', {{}}).get('percent') }}}}",
                "attributes_template": att_temp,
                "enabled": True,
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        devices = self._get_devices()
        new_devices = set(devices) - set(self.devices)
        disconnected_devices = set(self.devices) - set(devices)
        for device_name in disconnected_devices:
            devices[device_name] = self.devices[device_name]
            self.devices[device_name]["percent"] = None
        self.devices = devices
        if len(new_devices) > 0:
            self.lnxlink.setup_discovery()

        return devices

    def _get_devices(self):
        devices = {}
        for device in self.get_batteries():
            native_path = device["NativePath"].split("/")[-1]
            name = " ".join(
                [
                    device["Vendor"],
                    device["Model"],
                    device["Serial"].replace(":", ""),
                ]
            ).strip()
            if name == "":
                name = native_path
            if name != "":
                devices[name] = {
                    "percent": device["Percentage"],
                    "attributes": {
                        "vendor": device["Vendor"],
                        "model": device["Model"],
                        "serial": device["Serial"],
                        "native_path": native_path,
                        "rechargeable": device["IsRechargeable"],
                    },
                }
        return devices

    def dbus_paths(self, service, object_path, paths):
        """Recursive method to read the list of paths from the service"""
        obj = self.bus.get_proxy(
            service, object_path, "org.freedesktop.DBus.Introspectable"
        )
        xml_string = obj.Introspect()
        for child in ElementTree.fromstring(xml_string):
            if child.tag == "node":
                if object_path == "/":
                    object_path = ""
                new_path = "/".join((object_path, child.attrib["name"]))
                paths.append(new_path)
                self.dbus_paths(service, new_path, paths)
        return paths

    def get_batteries(self):
        """Gets a list of all devices and their status"""
        batteries = []
        for dbus_path in self.dbus_paths(
            "org.freedesktop.UPower", "/org/freedesktop/UPower", []
        ):
            proxy = self.bus.get_proxy(
                service_name="org.freedesktop.UPower",
                object_path=dbus_path,
                interface_name="org.freedesktop.UPower.Device",
            )
            # pylint: disable=protected-access
            if (
                "org.freedesktop.UPower.Device"
                in proxy._handler.specification.interfaces
            ):
                if proxy.Model + proxy.NativePath != "":
                    batteries.append(
                        {
                            "Model": proxy.Model,
                            "NativePath": proxy.NativePath,
                            "Percentage": proxy.Percentage,
                            "Serial": proxy.Serial,
                            "IconName": proxy.IconName,
                            "IsRechargeable": proxy.IsRechargeable,
                            "Vendor": proxy.Vendor,
                        }
                    )
        return batteries



"""Gets WiFi information"""
import os
from lnxlink.modules.scripts.helpers import import_install_package


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "WiFi"
        if not os.path.exists("/proc/net/wireless"):
            raise SystemError("No WiFi found.")
        self._requirements()

    def _requirements(self):
        self.lib = {
            "dbus_nd": import_install_package(
                "dbus-networkdevices", ">=2024.0.7", "dbus_networkdevices"
            ),
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "WiFi": {
                "type": "sensor",
                "icon": "mdi:wifi",
                "unit": "%",
                "entity_category": "diagnostic",
                "state_class": "measurement",
                "value_template": "{{ value_json.signal }}",
                "attributes_template": "{{ value_json.attributes | tojson }}",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        interface = ""
        ssid = ""
        mac = ""
        signal = None
        devices = self.lib["dbus_nd"].DBUSNetworkDevices().get_network_devices()
        for device in devices:
            if "wifi" in device:
                interface = device["interface"]
                signal = device["wifi"]["strength"]
                ssid = device["wifi"]["ssid"]
                mac = device["wifi"]["mac"]

        return {
            "signal": signal,
            "attributes": {
                "Interface": interface,
                "SSID": ssid,
                "MAC": mac,
            },
        }



"""Checks GPIO pins from a raspberry"""
import os
import time
from collections import deque
from threading import Thread
from lnxlink.modules.scripts.helpers import import_install_package


class GpioHandle:
    """Handle the Raspberry PI GPIO inputs by fixing spike events"""

    def __init__(self, gpio, pin, callback, setup="input"):
        """Start checking for pin values in a new thread"""
        self.gpio = gpio
        self.pin = pin
        self.callback = callback
        self.setup = setup
        self.pinvalue = None
        readgpio_thr = Thread(target=self.read, daemon=True)
        readgpio_thr.start()

    def read(self):
        """Gets the average value of the input pin and send it to the callback"""
        if self.setup == "input":
            self.gpio.setup(self.pin, self.gpio.IN, pull_up_down=self.gpio.PUD_UP)
            deq = deque(maxlen=15)
        elif self.setup == "output":
            deq = deque(maxlen=1)
        while True:
            if self.setup == "input":
                deq.append(self.gpio.input(self.pin))
            else:
                deq.append(0)
                if self.gpio.gpio_function(self.pin) == self.gpio.OUT:
                    deq.append(1)
            pinvalue = round(sum(deq) / deq.maxlen)
            if pinvalue != self.pinvalue:
                self.pinvalue = pinvalue
                self.callback(self.pin, self.pinvalue, self.setup)
            time.sleep(0.1)


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "gpio"
        self.lnxlink = lnxlink
        self.gpio_results = {}
        self.started = False
        if not self._is_raspberry():
            raise SystemError("Not supported non Raspberry PI devices")
        self._requirements()

    def get_info(self):
        """Starts only once the GPIO class for each pin"""
        if self._is_raspberry() and not self.started:
            self.started = True
            for device in self.lnxlink.config["settings"]["gpio"]["inputs"]:
                GpioHandle(
                    self.lib["gpio"].GPIO, device["pin"], self.pin_callback, "input"
                )
            for device in self.lnxlink.config["settings"]["gpio"]["outputs"]:
                GpioHandle(
                    self.lib["gpio"].GPIO, device["pin"], self.pin_callback, "output"
                )

    def pin_callback(self, pin, pinvalue, pintype):
        """Sends the data to the MQTT broker asyncronous"""
        value = "ON"
        if pinvalue == 0:
            value = "OFF"
        filter_input_gpio = list(
            filter(
                lambda pins: pins["pin"] == pin,
                self.lnxlink.config["settings"]["gpio"]["inputs"],
            )
        )
        filter_output_gpio = list(
            filter(
                lambda pins: pins["pin"] == pin,
                self.lnxlink.config["settings"]["gpio"]["outputs"],
            )
        )
        filter_gpio = filter_input_gpio + filter_output_gpio
        self.gpio_results[f"{pintype}_{filter_gpio[0]['name']}"] = value
        self.lnxlink.run_module(self.name, self.gpio_results)

    def _is_raspberry(self):
        model = ""
        if os.path.exists("/proc/device-tree/model"):
            with open("/proc/device-tree/model", "r", encoding="UTF-8") as f:
                model = f.read()
        if "raspberry" in model.lower():
            return True
        return False

    def _requirements(self):
        self.lib = {
            "gpio": import_install_package("RPi.GPIO"),
        }
        self.lib["gpio"].GPIO.setmode(self.lib["gpio"].GPIO.BCM)

    def exposed_controls(self):
        """Exposes to home assistant"""
        if not self._is_raspberry():
            return {}
        discovery_info = {}
        for device in self.lnxlink.config["settings"]["gpio"]["inputs"]:
            discovery_info[f"GPIO {device['name']}"] = {
                "type": "binary_sensor",
                "value_template": f"{{{{ value_json.get('input_{device['name']}') }}}}",
                "device_class": device.get("device_class", "None"),
            }
        for device in self.lnxlink.config["settings"]["gpio"]["outputs"]:
            discovery_info[f"GPIO {device['name']}"] = {
                "type": "switch",
                "value_template": f"{{{{ value_json.get('output_{device['name']}') }}}}",
                "icon": device.get("device_class", "mdi:bullhorn"),
            }
        return discovery_info

    def start_control(self, topic, data):
        """Control system"""
        outputs = self.lnxlink.config["settings"]["gpio"]["outputs"]
        outputs = {
            item["name"].lower().replace(" ", "_"): item["pin"] for item in outputs
        }

        out_key = topic[1].replace("gpio_", "")
        pin = outputs[out_key]
        if data.lower() == "on":
            self.lib["gpio"].GPIO.setup(pin, self.lib["gpio"].GPIO.OUT)
            self.lib["gpio"].GPIO.output(pin, self.lib["gpio"].GPIO.HIGH)
        elif data.lower() == "off":
            self.lib["gpio"].GPIO.setup(pin, self.lib["gpio"].GPIO.OUT)
            self.lib["gpio"].GPIO.output(pin, self.lib["gpio"].GPIO.LOW)
            self.lib["gpio"].GPIO.setup(pin, self.lib["gpio"].GPIO.IN)



"""Gets Inference Time information"""


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Inference Time"
        self.lnxlink = lnxlink

    def get_info(self):
        """Gather information from the system"""
        return {
            "modules": self.lnxlink.inference_times,
            "sum": round(sum(self.lnxlink.inference_times.values()), 2),
            "max": max(
                self.lnxlink.inference_times, key=self.lnxlink.inference_times.get
            ),
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Inference Time": {
                "type": "sensor",
                "icon": "mdi:timelapse",
                "unit": "s",
                "entity_category": "diagnostic",
                "state_class": "measurement",
                "value_template": "{{ value_json.sum }}",
                "attributes_template": "{{ value_json.modules | tojson }}",
            },
            "Inference Time Highest": {
                "type": "sensor",
                "icon": "mdi:timelapse",
                "entity_category": "diagnostic",
                "value_template": "{{ value_json.max }}",
                "enabled": False,
            },
        }



"""Shows notifications"""
import logging
import tempfile
import requests
from lnxlink.modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Notify"
        self.lnxlink = lnxlink
        self._requirements()
        self.notify = self.lib["notify"].DBusNotification(
            appname="LNXlink", callback=self.callback_action
        )

    def _requirements(self):
        self.lib = {
            "notify": import_install_package(
                "dbus-notification", ">=2024.7.2", "dbus_notification"
            ),
        }

    # pylint: disable=too-many-locals, too-many-branches
    def start_control(self, topic, data):
        """Control system"""
        icon_url = data.get("iconUrl")
        sound_url = data.get("sound")
        icon_path = icon_url
        sound_path = sound_url
        if icon_url is not None and icon_url.startswith("http"):
            try:
                with tempfile.NamedTemporaryFile(
                    prefix="lnxlink_icon_", delete=False
                ) as icon_file:
                    img_data = requests.get(icon_url, timeout=3).content
                    icon_file.write(img_data)
                    icon_path = icon_file.name
            except Exception as err:
                logger.error("Error downloading notification image: %s", err)
        if sound_url is not None and sound_url.startswith("http"):
            try:
                with tempfile.NamedTemporaryFile(
                    prefix="lnxlink_sound_", delete=False
                ) as sound_file:
                    sound_data = requests.get(sound_url, timeout=3).content
                    sound_file.write(sound_data)
                    sound_path = sound_file.name
            except Exception as err:
                logger.error("Error downloading notification sound: %s", err)

        urgencies = {
            "low": 0,
            "normal": 1,
            "critical": 2,
        }

        # notify2
        notification_id = self.notify.send(
            title=data["title"],
            message=data["message"],
            logo=f"{self.lnxlink.path}/logo.png",
            image=icon_path,
            sound=sound_path,
            actions=data.get("buttons", []),
            urgency=urgencies.get(data.get("urgency")),
            timeout=data.get("timeout", -1),
        )
        logger.debug("The notification %s was sent.", notification_id)

    def callback_action(self, notification_type, notification):
        """Gather notification options and send to the MQTT broker"""
        if notification_type == "button":
            logger.debug("Pressed notification button: %s", notification)
            self.lnxlink.run_module(f"{self.name}/button_press", notification)



"""Checks if the microphone is used"""
from importlib import import_module
from threading import Thread
import logging
from lnxlink.modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Speech Recognition"
        self.run = False
        self.speech = ""
        try:
            import_module("pyaudio")
        except Exception as err:
            raise SystemError(
                "System package 'python3-pyaudio' or 'portaudio19-dev' not installed"
            ) from err
        self._requirements()

    def _requirements(self):
        self.lib = {
            "alsaaudio": import_install_package("pyalsaaudio", ">=0.9.2", "alsaaudio"),
            "sr": import_install_package(
                "SpeechRecognition", ">=3.10.0", "speech_recognition"
            ),
        }

    def get_info(self):
        """Gather information from the system"""
        status = "OFF"
        if self.run:
            status = "ON"
        return {
            "status": status,
            "speech": self.speech,
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Speech Recognition": {
                "type": "binary_sensor",
                "icon": "mdi:account-tie-voice",
                "value_template": "{{ value_json.status }}",
            },
            "Start Speech Recognition": {
                "type": "button",
                "icon": "mdi:account-tie-voice",
            },
        }

    def start_recognition(self):
        """Start a voice recognition"""
        try:
            recognizer = self.lib["sr"].Recognizer()
            with self.lib["sr"].Microphone() as source:
                audio = recognizer.listen(source, timeout=2, phrase_time_limit=3)
                self.speech = recognizer.recognize_google(audio)
        except Exception as err:
            logger.error("Error with speech recognition: %s", err)
            self.speech = ""
        self.run = False

    def start_control(self, topic, data):
        """Control system"""
        if not self.run:
            self.run = True
            background = Thread(target=self.start_recognition)
            background.start()



"""Gets CPU usage information"""
import psutil
from lnxlink.modules.scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "CPU Usage"
        self.lnxlink = lnxlink
        self.cpuinfo = self._cpuinfo()

    def get_info(self):
        """Gather information from the system"""
        return {
            "percent": psutil.cpu_percent(),
            "attributes": {"CPU Info": self.cpuinfo},
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        update_interval = self.lnxlink.config.get("update_interval", 5)
        return {
            "CPU Usage": {
                "type": "sensor",
                "icon": "mdi:speedometer",
                "unit": "%",
                "state_class": "measurement",
                "expire_after": update_interval * 5,
                "value_template": "{{ value_json.get('percent')}}",
                "attributes_template": "{{ value_json.get('attributes', {}) | tojson }}",
            },
        }

    def _cpuinfo(self):
        cmd = (
            "cat /proc/cpuinfo | grep -i 'model name' | uniq | awk -F ':' '{print $2}'"
        )
        stdout, _, _ = syscommand(cmd)
        return stdout



"""Checks if the webcam is used"""
import glob
from threading import Thread
from inotify.adapters import Inotify


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Camera used"
        self.lnxlink = lnxlink
        self.inotify = Inotify()
        self.cameras = []
        Thread(target=self._watch_events, daemon=True).start()

    def get_info(self):
        """Gather information from the system"""
        cameras = glob.glob("/dev/video*", recursive=True)
        if cameras != self.cameras:
            self.cameras = cameras
            for camera in cameras:
                self.inotify.add_watch(camera)

    def _watch_events(self):
        for event in self.inotify.event_gen(yield_nones=False):
            cam_used = event[1][0] == "IN_OPEN"
            self.lnxlink.run_module(self.name, cam_used)

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Camera used": {
                "type": "binary_sensor",
                "icon": "mdi:webcam",
            },
        }



"""Checks for LNXlink updates"""
import re
import sys
import logging
import time
import requests
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "LNXlink update"
        self.lnxlink = lnxlink
        self.last_time = 0
        self.update_interval = 86400  # Check for updates every 24 hours
        self.message = {
            "installed_version": self.lnxlink.version,
            "latest_version": self.lnxlink.version,
            "release_summary": "",
            "release_url": "https://github.com/bkbilly/lnxlink/releases/latest",
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Update": {
                "type": "update",
                "title": "LNXlink",
                "icon": "mdi:update",
                "entity_category": "diagnostic",
                "entity_picture": "https://github.com/bkbilly/lnxlink/raw/master/logo.png?raw=true",
                "install": "install",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        cur_time = time.time()
        if cur_time - self.last_time > self.update_interval:
            self._latest_version()
            self.last_time = cur_time

        return self.message

    def _latest_version(self):
        """Gets the currently published version of lnxlink"""
        url = "https://api.github.com/repos/bkbilly/lnxlink/releases/latest"
        try:
            resp = requests.get(url=url, timeout=5).json()
            body = re.sub(r"##.*\n", "", resp["body"])
            self.message["latest_version"] = resp["tag_name"]
            self.message["release_summary"] = body
            self.message["release_url"] = resp["html_url"]
        except Exception as err:
            logger.error(err)

    def start_control(self, topic, data):
        """Control system"""
        if "+edit" in self.lnxlink.version:
            syscommand(f"git -C {self.lnxlink.path} pull", timeout=15)
            syscommand(
                f"{sys.executable} -m pip install -e {self.lnxlink.path}", timeout=120
            )
        else:
            syscommand(f"{sys.executable} -m pip install -U lnxlink", timeout=120)
        self.lnxlink.restart_script()



"""Checks for system updates"""
import time
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "System Updates"
        self.last_time = 0
        self.update_interval = 360  # Check for updates every 6 minutes
        self.updates = {
            "needs_update": "OFF",
            "packages": {"updates": []},
        }
        self.package_manager = None
        if which("apt") is not None:
            self.package_manager = {
                "command": "apt list --upgradable | grep -v 'Listing...' | awk -F '/' '{print $1}'",
                "largerthan": 0,
            }
        elif which("yum") is not None:
            self.package_manager = {
                "command": "yum -q updateinfo list updates",
                "largerthan": 2,
            }
        elif which("pacman") is not None:
            self.package_manager = {
                "command": "pacman -Qu",
                "largerthan": 0,
            }
        elif which("dnf") is not None:
            self.package_manager = {
                "command": "dnf updateinfo list",
                "largerthan": 1,
            }
        else:
            raise SystemError("System commands not found for package manager")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "System Updates": {
                "type": "binary_sensor",
                "icon": "mdi:package-variant",
                "entity_category": "diagnostic",
                "value_template": "{{ value_json.needs_update }}",
            },
            "System Updates Count": {
                "type": "sensor",
                "icon": "mdi:package-variant",
                "entity_category": "diagnostic",
                "value_template": "{{ value_json.packages.updates | count }}",
                "attributes_template": "{{ value_json.packages | tojson }}",
                "enabled": False,
            },
        }

    def get_info(self):
        """Gather information from the system"""
        cur_time = time.time()
        if cur_time - self.last_time > self.update_interval:
            self.last_time = cur_time
            stdout, _, _ = syscommand(self.package_manager["command"])
            if len(stdout) == 0:
                self.updates["needs_update"] = "OFF"
                self.updates["packages"]["updates"] = []
            else:
                current_updates = stdout.split("\n")
                needs_update = len(current_updates) > self.package_manager["largerthan"]
                self.updates["needs_update"] = "ON" if needs_update else "OFF"
                self.updates["packages"]["updates"] = current_updates
        return self.updates



"""Keeps display on"""
import re
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.lnxlink = lnxlink
        self.name = "Keep Alive"
        self.keepalive = "OFF"
        if which("gsettings") is None or which("xset") is None:
            raise SystemError("System commands 'gsettings' or 'xset' not found")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Keep Alive": {
                "type": "switch",
                "icon": "mdi:mouse-variant",
            }
        }

    def get_info(self):
        """Gather information from the system"""
        enabled_list = []

        # Check if Gnome Idle Time is active
        stdout_dim, _, _ = syscommand(
            "gsettings get org.gnome.desktop.session idle-delay"
        )
        stdout_suspend, _, _ = syscommand(
            "gsettings get org.gnome.settings-daemon.plugins.power sleep-inactive-ac-type"
        )
        if stdout_dim != "":
            disabled_dim = "0" == stdout_dim.split()[1]
            if disabled_dim and stdout_suspend != "":
                enabled_list.append("nothing" in stdout_suspend)

        # Check if DPMS is active
        if self.lnxlink.display is not None:
            stdout_xset, _, _ = syscommand(f"xset -display {self.lnxlink.display} q")
            xset_pattern = re.compile(r"Standby: (\d+)\s+Suspend: (\d+)\s+Off: (\d+)")
            xset_match = re.findall(xset_pattern, stdout_xset)
            for nums in xset_match:
                enabled_list.append(all(num != "0" for num in nums))
                if enabled_list[-1]:
                    enabled_list.append("DPMS is Enabled" in stdout_xset)

        return any(enabled_list)

    def start_control(self, topic, data):
        """Control system"""
        if data.lower() == "off":
            syscommand(
                "gsettings set org.gnome.desktop.session idle-delay 600",
            )
            syscommand(
                'gsettings set org.gnome.settings-daemon.plugins.power sleep-inactive-ac-type "suspend"',
            )
            syscommand(
                "xset +dpms",
            )
        elif data.lower() == "on":
            syscommand(
                "gsettings set org.gnome.desktop.session idle-delay 0",
            )
            syscommand(
                'gsettings set org.gnome.settings-daemon.plugins.power sleep-inactive-ac-type "nothing"',
            )
            syscommand(
                "xset -dpms",
            )



"""A collection of helper functions"""
import sys
import importlib.metadata
import logging
import subprocess

logger = logging.getLogger("lnxlink")


# pylint: disable=consider-using-with
def syscommand(command, ignore_errors=False, timeout=3, background=False):
    """Global subprocess command"""
    if isinstance(command, list):
        command = " ".join(command)
    if background:
        subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return "", "", 0
    result = subprocess.run(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        timeout=timeout,
    )
    stderr = result.stderr.decode("UTF-8").strip()
    stdout = result.stdout.decode("UTF-8").strip()
    returncode = result.returncode
    if returncode != 0 and ignore_errors is False:
        logger.error("Error with command: %s (%s)", command, stderr)
    return stdout, stderr, returncode


def import_install_package(package, req_version="", syspackage=None):
    """Imports a system package and if it doesn't exist, it gets installed"""
    if syspackage is None:
        syspackage = package

    try:
        current_version = importlib.metadata.version(package)
    except importlib.metadata.PackageNotFoundError:
        current_version = None

    if current_version is None or needs_update(current_version, req_version):
        package_version = f"'{package}{req_version}'"
        logger.info("Installing %s...", package_version)
        args = [
            sys.executable,
            "-m",
            "pip",
            "install",
            "--break-system-packages",
            "-U",
            "--quiet",
            package_version,
        ]
        _, _, returncode = syscommand(args, ignore_errors=True, timeout=None)
        if returncode != 0:
            try:
                return __import__(syspackage)
            except ModuleNotFoundError:
                logger.error("Can't install package %s", package)
                return None

    try:
        return __import__(syspackage)
    except Exception as err:
        logger.error("Can't import package %s: %s", package, err)
        return None


def needs_update(current_version, request_version):
    """Compares two version strings"""
    current_version = str(current_version).strip("><=~")
    request_version = str(request_version).strip("><=~")
    try:
        current_versions = [int(version) for version in current_version.split(".")]
    except (TypeError, ValueError):
        current_versions = [0]
    if request_version is None or request_version == "":
        return False
    request_versions = [int(version) for version in request_version.split(".")]
    for num in range(max(len(current_versions), len(request_versions))):
        cur_version = current_versions[num] if num < len(current_versions) else 0
        req_version = request_versions[num] if num < len(request_versions) else 0
        if req_version > cur_version:
            return True
        if req_version < cur_version:
            return False
    return False



"""Monitor Gamepad controllers for button presses"""
import re
import time
import struct
import logging
from threading import Thread
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Read events by Gamepad"""

    def __init__(self, lnxlink):
        self.name = "Gamepad"
        self.gamepads = []
        self.running_threads = []
        self.last_used = 0
        self.timeout_used = 40

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Gamepad Used": {
                "type": "binary_sensor",
                "icon": "mdi:controller",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        self.watch_gamepads()
        if int(time.time()) - self.last_used < self.timeout_used:
            return True
        return False

    def watch_gamepads(self):
        """Watch for gamepad connections"""
        stdout, _, _ = syscommand(
            "cat /proc/bus/input/devices | grep -P '^H:.* js[0-9]+'", ignore_errors=True
        )
        match = re.findall(r"(event\d+)", stdout)
        if self.gamepads != match:
            logger.info("Gamepads found: %s", match)
            self.gamepads = match
            for running_thread in self.running_threads:
                running_thread.join(1)
            self.running_threads = []
            for event in match:
                watch_thr = Thread(target=self.watch_input, args=(event,), daemon=True)
                watch_thr.start()
                logger.debug("Started for: %s", event)
                self.running_threads.append(watch_thr)

    def watch_input(self, event):
        """Thread that watches gamepad inputs"""
        decode_str = "llHHI"
        with open(f"/dev/input/{event}", "rb") as file:
            while game_data := file.read(struct.calcsize(decode_str)):
                _, _, ev_type, code, value = struct.unpack(decode_str, game_data)
                if ev_type != 0 or code != 0 or value != 0:
                    self.last_used = int(time.time())
                    logger.debug(code, value)



"""Gets the network usage"""
from datetime import datetime
import psutil


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Network"
        self.time_old = datetime.now()
        self.recv_old = psutil.net_io_counters().bytes_recv
        self.sent_old = psutil.net_io_counters().bytes_sent

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Network Upload": {
                "type": "sensor",
                "icon": "mdi:access-point-network",
                "unit": "Mbit/s",
                "state_class": "measurement",
                "device_class": "data_rate",
                "value_template": "{{ value_json.upload }}",
            },
            "Network Download": {
                "type": "sensor",
                "icon": "mdi:access-point-network",
                "unit": "Mbit/s",
                "state_class": "measurement",
                "device_class": "data_rate",
                "value_template": "{{ value_json.download }}",
            },
        }

    def get_info(self):
        """Returns Mbps"""
        time_new = datetime.now()
        netio = psutil.net_io_counters()
        recv_new = netio.bytes_recv
        sent_new = netio.bytes_sent

        time_diff = (time_new - self.time_old).total_seconds()
        self.time_old = time_new

        recv_diff = recv_new - self.recv_old
        sent_diff = sent_new - self.sent_old
        self.recv_old = recv_new
        self.sent_old = sent_new

        recv_speed = max(0, round(recv_diff * 8 / time_diff / 1024 / 1024, 2))
        sent_speed = max(0, round(sent_diff * 8 / time_diff / 1024 / 1024, 2))

        return {
            "upload": sent_speed,
            "download": recv_speed,
        }



"""Controls Paired Bluetooth Devices"""
import re
import logging
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Bluetooth"
        self.lnxlink = lnxlink
        _, _, returncode = syscommand("bluetoothctl show")
        if returncode != 0:
            raise SystemError("Can't run command bluetoothctl")
        self.bluetoothdata = self._get_bluetoothdata()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {
            "Bluetooth Power": {
                "type": "switch",
                "icon": "mdi:bluetooth-connect",
                "value_template": "{{ value_json.get('power') }}",
            }
        }
        for mac, blinfo in self.bluetoothdata["devices"].items():
            discovery_info[
                f"Bluetooth Device {blinfo['name']} {mac.replace(':', '')}"
            ] = {
                "type": "switch",
                "icon": "mdi:bluetooth",
                "value_template": f"{{{{ value_json.devices.get('{mac}', {{}}).get('power') }}}}",
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        bluetoothdata = self._get_bluetoothdata()
        loaded = set(bluetoothdata["devices"]) - set(self.bluetoothdata["devices"])
        unloaded = set(self.bluetoothdata["devices"]) - set(bluetoothdata["devices"])
        for device in unloaded:
            for addr in self.bluetoothdata[device]:
                self.bluetoothdata[device][addr] = None
            bluetoothdata[device] = self.bluetoothdata[device]
        self.bluetoothdata = bluetoothdata
        if len(loaded) > 0:
            self.lnxlink.setup_discovery()
        return self.bluetoothdata

    def _bytetogb(self, byte):
        return round(byte / 1024 / 1024 / 1024, 0)

    def _get_bluetoothdata(self):
        """Get a list of all bluetooth devices"""
        data = {
            "power": "OFF",
            "devices": {},
        }
        stdout, _, _ = syscommand("bluetoothctl show | grep Powered")
        if "yes" in stdout:
            data["power"] = "ON"

        stdout, _, returncode = syscommand("bluetoothctl devices Paired")
        if returncode != 0:
            stdout, _, _ = syscommand("bluetoothctl paired-devices")
        if stdout != "":
            for device in stdout.split("\n"):
                _, mac, name = device.split(" ", maxsplit=2)
                stdoutdevice, _, _ = syscommand(
                    f"bluetoothctl info {mac} | grep Connected"
                )
                power = "OFF"
                if "yes" in stdoutdevice:
                    power = "ON"
                data["devices"][mac] = {
                    "name": name,
                    "power": power,
                }

        return data

    def start_control(self, topic, data):
        """Control system"""
        if topic[1] == "bluetooth_power":
            syscommand(f"bluetoothctl power {data.lower()}")
        else:
            simple_mac = topic[1].split("_")[-1]
            mac = ":".join(re.findall("..?", simple_mac))
            if data == "ON":
                syscommand(f"bluetoothctl connect {mac}")
            elif data == "OFF":
                syscommand(f"bluetoothctl disconnect {mac}")



"""Controls and decodes IR signals on a raspberry"""
import os
import json
import time
import logging
from threading import Thread
from lnxlink.modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "IR Remote"
        self.lnxlink = lnxlink
        self.gpio_results = {}
        self.started = False
        if not self._is_raspberry():
            raise SystemError("Not supported non Raspberry PI devices")
        self._requirements()
        self.irremote = None

    def get_info(self):
        """Starts only once the GPIO class for each pin"""
        if not self.started:
            self.irremote = IRRemote(self.lib["pigpio"])
            self.started = True
            receiver = self.lnxlink.config["settings"]["ir_remote"]["receiver"]
            if receiver:
                self.irremote.setup_receiver(receiver, self.receiver_callback)

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        receiver = self.lnxlink.config["settings"]["ir_remote"]["receiver"]
        transmitter = self.lnxlink.config["settings"]["ir_remote"]["transmitter"]
        if receiver:
            discovery_info["IR Receiver"] = {
                "type": "sensor",
                "icon": "mdi:square-wave",
                "value_template": "{{ value_json.decsignal }}",
                "subtopic": True,
            }
        if transmitter:
            discovery_info["IR Transmitter"] = {
                "type": "text",
                "icon": "mdi:square-wave",
                "max": 1000,
            }
            for button in self.lnxlink.config["settings"]["ir_remote"]["buttons"]:
                discovery_info[f"IR {button['name']}"] = {
                    "type": "button",
                    "icon": button.get("icon", "mdi:square-wave"),
                    "payload_press": json.dumps(button["data"]),
                }

        return discovery_info

    def start_control(self, topic, data):
        """Control system"""
        transmitter = self.lnxlink.config["settings"]["ir_remote"]["transmitter"]
        if transmitter:
            self.irremote.pause = True
            self.irremote.send_signal(transmitter, data)
            self.irremote.pause = False

    def receiver_callback(self, rawsignal, binsignal, decsignal, protocol):
        """Callback that sends the data to the MQTT broker"""
        tosend = {
            "rawsignal": rawsignal,
            "binsignal": binsignal,
            "decsignal": decsignal,
            "protocol": protocol,
        }
        self.lnxlink.run_module(f"{self.name}/IR Receiver", tosend)
        logger.debug("{%s}:{%s} = {%s}", protocol, decsignal, binsignal)

    def _requirements(self):
        self.lib = {
            "pigpio": import_install_package("pigpio"),
        }

    def _is_raspberry(self):
        model = ""
        if os.path.exists("/proc/device-tree/model"):
            with open("/proc/device-tree/model", "r", encoding="UTF-8") as f:
                model = f.read()
        if "raspberry" in model.lower():
            return True
        return False


class IRRemote:
    """docstring for IRRemote"""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, pigpio):
        self.pigpio = pigpio
        self.pi = self.pigpio.pi()  # Connect to Pi.
        if not self.pi.connected:
            raise SystemError("Make sure the pigpiod daemon is running")

        self.gpio_rx = None
        self.last_tick = 0
        self.in_code = False
        self.ir_signal = []
        self.fetching_code = False

        self.read_thr = None
        self.pause = False

    def start_receiving(self, callback):
        """Start listening for IR commands"""
        while True:
            self.ir_signal = []
            self.fetching_code = True
            while self.fetching_code:
                time.sleep(0.1)
            time.sleep(0.5)
            binsignal, decsignal, protocol = SignalDecoder().decode_signal(
                self.ir_signal
            )
            callback(self.ir_signal, binsignal, decsignal, protocol)
            time.sleep(0.1)

    def setup_receiver(self, gpio, callback):
        """Setup GPIO and it's callback"""
        self.gpio_rx = gpio
        if self.gpio_rx:
            glitch = 100
            self.pi.set_mode(
                self.gpio_rx, self.pigpio.INPUT
            )  # IR RX connected to this GPIO.
            self.pi.set_glitch_filter(self.gpio_rx, glitch)  # Ignore glitches.
            self.pi.callback(self.gpio_rx, self.pigpio.EITHER_EDGE, self.cbf)
            if self.read_thr is None:
                self.read_thr = Thread(
                    target=self.start_receiving, args=(callback,), daemon=True
                )
                self.read_thr.start()

    def disconnect(self):
        """Disconnects from GPIO and stops any running process"""
        if self.gpio_rx:
            self.pi.set_glitch_filter(self.gpio_rx, 0)  # Cancel glitch filter.
            self.pi.set_watchdog(self.gpio_rx, 0)  # Cancel watchdog.
            self.pi.stop()  # Disconnect from Pi
            if self.read_thr is not None:
                self.read_thr.join()
                self.read_thr = None

    def send_signal(self, gpio, ir_signal):
        """Sends an IR Signal to the IR LED"""
        self.pi.set_mode(gpio, self.pigpio.OUTPUT)  # IR TX connected to this GPIO.

        # Create wave
        self.pi.wave_add_new()
        waves_dict = {}
        wave = []
        for num, ci in enumerate(ir_signal):
            ci = int(ci)
            if num & 1:  # Space
                if ci not in waves_dict:
                    self.pi.wave_add_generic([self.pigpio.pulse(0, 0, ci)])
                    waves_dict[ci] = self.pi.wave_create()
            else:  # Pulse
                if ci not in waves_dict:
                    wf = self.carrier(gpio, ci)
                    self.pi.wave_add_generic(wf)
                    waves_dict[ci] = self.pi.wave_create()
            wave.append(waves_dict[ci])

        self.pi.wave_chain(wave)
        while self.pi.wave_tx_busy():
            time.sleep(0.002)
        self.pi.wave_clear()
        self.pi.set_mode(gpio, self.pigpio.INPUT)

    def carrier(self, gpio, micros):
        """Generate carrier square wave"""
        frequency = 38.0
        wf = []
        cycle = 1000.0 / frequency
        cycles = int(round(micros / cycle))
        on = int(round(cycle / 2.0))
        sofar = 0
        for c in range(cycles):
            target = int(round((c + 1) * cycle))
            sofar += on
            off = target - sofar
            sofar += off
            wf.append(self.pigpio.pulse(1 << gpio, 0, on))
            wf.append(self.pigpio.pulse(0, 1 << gpio, off))
        return wf

    def normalise(self, c):
        """This function identifies the distinct pulses and takes the
        average of the lengths making up each distinct pulse.  Pulses
        and spaces are processed separately."""

        tolerance = 15
        toler_min = (100 - tolerance) / 100.0
        toler_max = (100 + tolerance) / 100.0

        entries = len(c)
        p = [0] * entries  # Set all entries not processed.
        for i in range(entries):
            if not p[i]:  # Not processed?
                v = c[i]
                tot = v
                similar = 1.0

                # Find all pulses with similar lengths to the start pulse.
                for j in range(i + 2, entries, 2):
                    if not p[j]:  # Unprocessed.
                        if (c[j] * toler_min) < v < (c[j] * toler_max):  # Similar.
                            tot = tot + c[j]
                            similar += 1.0

                # Calculate the average pulse length.
                newv = int(round(tot / similar, 0))
                c[i] = newv

                # Set all similar pulses to the average value.
                for j in range(i + 2, entries, 2):
                    if not p[j]:  # Unprocessed.
                        if (c[j] * toler_min) < v < (c[j] * toler_max):  # Similar.
                            c[j] = newv
                            p[j] = 1

    def end_of_code(self):
        """End of the IR code"""
        short = 10
        if len(self.ir_signal) > short:
            self.normalise(self.ir_signal)
            self.fetching_code = False
        else:
            self.ir_signal = []

    def cbf(self, gpio, level, tick):
        """Callback function of GPIO input"""
        pre_ms = 200
        post_ms = 25
        pre_us = pre_ms * 1000
        post_us = post_ms * 1000

        if level != self.pigpio.TIMEOUT:
            edge = self.pigpio.tickDiff(self.last_tick, tick)
            self.last_tick = tick

            if self.fetching_code:
                if (edge > pre_us) and (not self.in_code):  # Start of a self.ir_signal.
                    self.in_code = True
                    self.pi.set_watchdog(gpio, post_ms)  # Start watchdog.

                elif (edge > post_us) and self.in_code:  # End of a self.ir_signal.
                    self.in_code = False
                    self.pi.set_watchdog(gpio, 0)  # Cancel watchdog.
                    self.end_of_code()

                elif self.in_code:
                    if not self.pause:
                        self.ir_signal.append(edge)

        else:
            self.pi.set_watchdog(gpio, 0)  # Cancel watchdog.
            if self.in_code:
                self.in_code = False
                self.end_of_code()


class SignalDecoder:
    """Decode IR signals using specific protocols"""

    def decode_signal(self, ir_signal):
        """Tries to find the correct signal protocol"""
        if binary_signal := self.decode_nec(ir_signal.copy()):
            return binary_signal, int(binary_signal, 2), "NEC"
        if binary_signal := self.decode_rc5(ir_signal.copy()):
            return binary_signal, int(binary_signal, 2), "RC5"
        if binary_signal := self.decode_sirc(ir_signal.copy()):
            return binary_signal, int(binary_signal, 2), "SIRC"
        if binary_signal := self.decode_philips(ir_signal.copy()):
            return binary_signal, int(binary_signal, 2), "PHILIPS"
        return None, None, None

    def decode_nec(self, ir_signal):
        """▁▁▆▆▆▆▆▆▁▁▁▁▁▁▁▆▆▁▁▆▆▁▁▁▁▆▆▁▁
        |lpulse|lspace| 0 |  1  | 0 |"""

        logger.debug("Decoding using NEC Protocol")
        # protocol_leading = 9000
        # protocol_leading_space = 4500
        protocol_pulse = 562.5
        protocol_pulse_high = 1687.5
        width = 500

        ir_signal = ir_signal[2:]
        if len(set(ir_signal[0::2])) != 1:
            return None

        mybin = []
        for pulse, space in zip(ir_signal[0::2], ir_signal[1::2]):
            if not protocol_pulse - width < pulse < protocol_pulse + width:
                return None
            if protocol_pulse - width < space < protocol_pulse + width:
                mybin.append("0")
            elif protocol_pulse_high - width < space < protocol_pulse_high + width:
                mybin.append("1")
            else:
                return None
        return "".join(mybin)

    # pylint: disable=too-many-branches
    def decode_rc5(self, ir_signal):
        """▁▁▁▆▆▆▁▁▁▆▆▆▆▆▆▁▁▁▆▆▆▁▁▁
        |  1  |  1  | 0  |  0  |"""
        logger.debug("Decoding using RC5 Protocol")
        if len(set(ir_signal[0::2])) == 1:
            return None
        while len(set(ir_signal[1::2])) > 2:
            ir_signal.remove(max(ir_signal[1::2]))

        protocol_pulse = min(ir_signal)
        width = protocol_pulse / 2

        tmp_signal = [0]
        for pulse, space in zip(ir_signal[0::2], ir_signal[1::2]):
            if protocol_pulse - width < pulse < protocol_pulse + width:
                tmp_signal.append(1)
            elif protocol_pulse * 2 - width < pulse < protocol_pulse * 2 + width:
                tmp_signal.append(1)
                tmp_signal.append(1)
            else:
                logger.debug("pulse: %s", pulse)
                return None

            if protocol_pulse - width < space < protocol_pulse + width:
                tmp_signal.append(0)
            elif protocol_pulse * 2 - width < space < protocol_pulse * 2 + width:
                tmp_signal.append(0)
                tmp_signal.append(0)

        tmp_signal.append(1)
        tmp_signal.append(0)
        mybin = []
        for first, second in zip(tmp_signal[0::2], tmp_signal[1::2]):
            if first == 0 and second == 1:
                mybin.append("1")
            elif first == 1 and second == 0:
                mybin.append("0")
            else:
                logger.debug("can't decode: %s, %s", first, second)
                return None

        if len(mybin) < 4:
            return None
        mybin = mybin[3:]
        return "".join(mybin)

    def decode_sirc(self, ir_signal):
        """▁▁▆▆▆▆▆▆▁▁▁▁▁▁▁▆▆▆▁▁▆▆▆▆▆▁▁▆▆▆▆▆▁▁
        |lpulse|lspace|  0  |  1  |   1   |"""

        logger.debug("Decoding using Sony SIRC Protocol")
        protocol_leading = 2400
        protocol_leading_space = 600
        protocol_pulse = 600
        protocol_pulse_high = 1200
        width = 300

        leading_pulse = ir_signal.pop(0)
        if not protocol_leading - width < leading_pulse < protocol_leading + width:
            logger.debug("leading_pulse: %s", leading_pulse)
            return None
        leading_space = ir_signal.pop(0)
        if (
            not protocol_leading_space - width
            < leading_space
            < protocol_leading_space + width
        ):
            logger.debug("leading_space: %s", leading_space)
            return None
        if len(set(ir_signal[1::2])) > 2:
            return None

        mybin = []
        for pulse, space in zip(ir_signal[0::2], ir_signal[1::2]):
            if protocol_pulse - width < pulse < protocol_pulse + width:
                mybin.append("0")
            elif protocol_pulse_high - width < pulse < protocol_pulse_high + width:
                mybin.append("1")
            else:
                logger.debug("pulse: %s", pulse)
                return None

            if not protocol_pulse - width < space < protocol_pulse + width:
                # Stop repeated signal
                logger.debug("SIRC space: %s", space)
                break

        if "1" not in mybin:
            return None
        return "".join(mybin)

    def decode_philips(self, ir_signal):
        """▁▁▆▆▆▆▆▆▁▁▁▁▁▁▁▆▆▁▁▆▆▆▁▁▁▆▆▆▁▁▁
           |lpulse|lspace| 0 |  1  |  1  |
        Custom protocol that decodes philips IR signals"""

        logger.debug("Decoding using Custom Protocol")
        low = 430
        high = 860
        width = 200

        ir_signal = ir_signal[2:]
        mybin = []
        for pulse, _ in zip(ir_signal[0::2], ir_signal[1::2]):
            if low - width < pulse < low + width:
                mybin.append("0")
            elif high - width < pulse < high + width:
                mybin.append("1")
            else:
                mybin.append("0")
                mybin.append("1")
        return "".join(mybin)



"""Gets Display Environment"""
from lnxlink.modules.scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Display Environment"
        self.lnxlink = lnxlink

    def get_info(self):
        """Gather information from the system"""
        display_var, _, _ = syscommand("echo $DISPLAY")
        if display_var:
            if display_var == "":
                display_var = None
            self.lnxlink.display = display_var
            return display_var
        other_displays, _, _ = syscommand(
            "sed -zn 's/^DISPLAY=//p' /proc/*/environ 2> /dev/null | LC_ALL=C sort -zu | tr '\\0' '\\n'"
        )
        other_displays = other_displays.split("\n")
        if len(other_displays) > 0:
            if other_displays[0] == "":
                other_displays[0] = None
            self.lnxlink.display = other_displays[0]
            return other_displays[0]
        return None

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Display Environment": {
                "type": "sensor",
                "icon": "mdi:panorama-variant-outline",
                "entity_category": "diagnostic",
                "enabled": False,
            }
        }



"""Constant variables to use"""

SERVICEHEADLESS = """[Unit]
Description=LNXlink
After=network-online.target

[Service]
Type=simple
Restart=always
RestartSec=5
User=root

ExecStart={exec_cmd}

[Install]
WantedBy=default.target
"""

SERVICEUSER = """[Unit]
Description=LNXlink
After=network-online.target multi-user.target graphical.target
PartOf=graphical-session.target

[Service]
Type=simple
Restart=always
RestartSec=5

ExecStart={exec_cmd}

[Install]
WantedBy=default.target
"""

CONFIGTEMP = """
mqtt:
  prefix: 'lnxlink'
  clientId: 'DESKTOP-Linux'
  server: '192.168.1.1'
  port: 1883
  auth:
    user: 'user'
    pass: 'pass'
    tls: false
    keyfile: ""
    certfile: ""
    ca_certs: ""
  discovery:
    enabled: true
  lwt:
    enabled: true
    qos: 1
    retain: false
update_interval: 5
update_on_change: false
hass_url:
hass_api:
modules:
custom_modules:
exclude:
  - audio_select
  - bash
  - battery
  - boot_select
  - brightness
  - fullscreen
  - gpio
  - gpu
  - idle
  - inference_time
  - ir_remote
  - keep_alive
  - keyboard_hotkeys
  - media
  - mouse
  - network
  - notify
  - power_profile
  - screen_onoff
  - screenshot
  - send_keys
  - speech_recognition
  - systemd
  - webcam
  - xdg_open
settings:
  systemd:
  gpio:
    inputs:
    outputs:
  hotkeys:
  disk_usage:
    include_disks: []
    exclude_disks: []
  statistics: "https://analyzer.bkbilly.workers.dev"
  bash:
    allow_any_command: False
    expose:
  mounts:
    autocheck: false
    directories: []
  ir_remote:
    receiver: null
    transmitter: null
    buttons: []
"""



"""Run a terminal command"""
import logging
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "bash"
        self.lnxlink = lnxlink
        self.discovery_info = {}

    def get_info(self):
        """Gather information from the system"""
        for expose_name, discovery in self.discovery_info.items():
            if discovery.get("type") == "sensor":
                stdout, _, returncode = syscommand(discovery["local_command"])
                if returncode == 0:
                    self.lnxlink.run_module(f"{self.name}/{expose_name}", stdout)
            elif discovery.get("type") == "binary_sensor":
                stdout, _, _ = syscommand(discovery["local_command"])
                status = stdout.lower() not in ["false", "no", "0", ""]
                senddata = {
                    "status": "ON" if status else "OFF",
                    "attributes": {"raw": stdout.split("\n")},
                }
                self.lnxlink.run_module(f"{self.name}/{expose_name}", senddata)
            elif discovery.get("type") == "switch":
                stdout, _, _ = syscommand(
                    discovery["local_command"], ignore_errors=True
                )
                status = stdout.lower() not in ["false", "no", "0", ""]
                self.lnxlink.run_module(f"{self.name}/{expose_name}", status)

    def exposed_controls(self):
        """Exposes to home assistant"""
        self.discovery_info = {}
        if self.lnxlink.config["settings"]["bash"]["allow_any_command"]:
            self.discovery_info["Bash Command"] = {
                "type": "text",
                "icon": "mdi:bash",
            }

        exposed = self.lnxlink.config["settings"]["bash"]["expose"]
        exposed = [] if exposed is None else exposed
        for expose in exposed:
            expose_type = expose.get("type", "button")
            expose_name = f"Bash {expose['name']}"
            icon = expose.get("icon", "mdi:script-text")
            if expose_type == "button":
                self.discovery_info[expose_name] = {
                    "type": expose_type,
                    "icon": icon,
                    "payload_press": expose["command"],
                }
            elif expose_type == "sensor":
                self.discovery_info[expose_name] = {
                    "type": expose_type,
                    "icon": icon,
                    "unit": expose.get("unit"),
                    "local_command": expose.get("command"),
                    "subtopic": True,
                }
            elif expose_type == "binary_sensor":
                self.discovery_info[expose_name] = {
                    "type": expose_type,
                    "icon": icon,
                    "value_template": "{{ value_json.status }}",
                    "attributes_template": "{{ value_json.attributes | tojson }}",
                    "local_command": expose.get("command"),
                    "subtopic": True,
                }
            elif expose_type == "switch":
                self.discovery_info[expose_name] = {
                    "type": expose_type,
                    "icon": icon,
                    "local_command": expose.get("command"),
                    "command_on": expose.get("command_on"),
                    "command_off": expose.get("command_off"),
                    "subtopic": True,
                }
            if expose.get("entity_category") in ["diagnostic", "config"]:
                self.discovery_info[expose_name]["entity_category"] = expose[
                    "entity_category"
                ]

        return self.discovery_info

    def start_control(self, topic, data):
        """Control system"""
        allow_any_command = self.lnxlink.config["settings"]["bash"]["allow_any_command"]
        if allow_any_command:
            stdout, _, _ = syscommand(data, timeout=120)
            return stdout
        exposed = self.lnxlink.config["settings"]["bash"]["expose"]
        exposed = [] if exposed is None else exposed
        for expose in exposed:
            command_list = [
                expose.get("command", "").strip(),
                expose.get("command_on", "").strip(),
                expose.get("command_off", "").strip(),
            ]
            if data.strip() in command_list:
                stdout, _, _ = syscommand(data, timeout=120)
                return stdout
        logger.error(
            "Check bash configuration option allow_any_command to run this command: %s",
            data,
        )
        return None



"""Gets memory usage information"""
import psutil


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Memory Usage"

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Memory Usage": {
                "type": "sensor",
                "icon": "mdi:memory",
                "unit": "%",
                "state_class": "measurement",
                "value_template": "{{ value_json.percent }}",
            },
            "Memory Used": {
                "type": "sensor",
                "icon": "mdi:memory",
                "unit": "MB",
                "state_class": "measurement",
                "device_class": "data_size",
                "value_template": "{{ value_json.used }}",
                "enabled": False,
            },
            "Memory Available": {
                "type": "sensor",
                "icon": "mdi:memory",
                "unit": "MB",
                "state_class": "measurement",
                "device_class": "data_size",
                "value_template": "{{ value_json.available }}",
                "enabled": False,
            },
        }

    def get_info(self):
        """Gather information from the system"""
        vmem = psutil.virtual_memory()
        return {
            "percent": round(vmem.percent, 0),
            "used": round(vmem.used / 1024**2, 0),
            "available": round(vmem.available / 1024**2, 0),
        }



"""Checks if a window is in fullscreen"""
from lnxlink.modules.scripts.helpers import import_install_package


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Fullscreen"
        self.lnxlink = lnxlink
        self._requirements()

    def _requirements(self):
        self.lib = {
            "ewmh": import_install_package("ewmh", ">=0.1.6"),
            "xlib": import_install_package("python-xlib", ">=0.33", "Xlib.display"),
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Fullscreen": {
                "type": "binary_sensor",
                "icon": "mdi:alert-octagon-outline",
                "value_template": "{{ value_json.is_fullscreen }}",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        data = {
            "is_fullscreen": "OFF",
            "window": "",
        }
        if self.lnxlink.display is None:
            return data
        display = self.lib["xlib"].display.Display(self.lnxlink.display)
        ewmh = self.lib["ewmh"].EWMH(_display=display)
        windows = ewmh.getClientList()
        for win in windows:
            state = ewmh.getWmState(win, True)
            name = ewmh.getWmName(win)
            if "_NET_WM_STATE_FULLSCREEN" in state:
                data["is_fullscreen"] = "ON"
                data["window"] = name.decode()

        return data



"""Assign hotkeys to run commands"""
import json
import logging
from datetime import datetime
from requests import post, get
from lnxlink.modules.scripts.helpers import import_install_package, syscommand


logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Keyboard Hotkeys"
        self.lnxlink = lnxlink
        self.started = False
        self.lib = {}
        use_hass = any("type" in v for v in self.lnxlink.config["settings"]["hotkeys"])
        if use_hass:
            self.hass_url = self.lnxlink.config["hass_url"]
            hass_api = self.lnxlink.config["hass_api"]
            self.hass_headers = {
                "Authorization": f"Bearer {hass_api}",
                "content-type": "application/json",
            }
            if hass_api is None or self.hass_url is None:
                raise SystemError(
                    "Configuration option 'hass_api' or 'hass_url' is not set"
                )
            if not self._test_connection():
                raise SystemError("Can't connect to Home Assistant")
        self._requirements()

    def exposed_controls(self):
        """Exposes to home assistant"""
        use_sensor = any(
            "type" not in v for v in self.lnxlink.config["settings"]["hotkeys"]
        )
        discovery_info = {}
        if use_sensor:
            discovery_info = {
                "Keyboard Hotkey": {
                    "type": "sensor",
                    "device_class": "timestamp",
                    "value_template": "{{ value_json.timevalue }}",
                    "attributes_template": "{{ value_json.attributes | tojson }}",
                },
            }
        return discovery_info

    def _requirements(self):
        self.lib = {
            "xlib_hotkeys": import_install_package(
                "xlib-hotkeys", ">=2024.3.0", "xlib_hotkeys"
            ),
        }

    def get_info(self):
        """Gather information from the system"""
        if not self.started and self.lnxlink.display is not None:
            hm = self.lib["xlib_hotkeys"].HotKeysManager(self.lnxlink.display)
            actions = self.lnxlink.config["settings"]["hotkeys"]
            for action in actions:
                action_key = action.get("key")
                hm.hotkeys[action_key] = lambda action=action: self._activate(action)
            hm.start()
            self.started = True

    def _activate(self, act):
        response = ""
        action = act.copy()
        key = action.pop("key")
        if "type" in action:
            action_type = action.pop("type")
            if action_type == "action":
                url_action = action.pop("service").replace(".", "/")
                url = f"{self.hass_url}/api/services/{url_action}"
                post(url, headers=self.hass_headers, data=json.dumps(action))
            elif action_type == "state":
                url = f"{self.hass_url}/api/states/light.myroom"
                response = get(url, headers=self.hass_headers).json()["state"]
            elif action_type in ["conversation", "popup"]:
                url = f"{self.hass_url}/api/conversation/process"
                if action_type == "conversation":
                    response = post(
                        url, headers=self.hass_headers, data=json.dumps(action)
                    )
                elif action_type == "popup":
                    stdout, _, _ = syscommand(
                        f"timeout 15s zenity --display='{self.lnxlink.display}' "
                        + "--entry --title='Home Assistant' --text='Sentence:'",
                        timeout=15,
                    )
                    if stdout == "":
                        return
                    action = {"text": stdout}
                    response = post(
                        url, headers=self.hass_headers, data=json.dumps(action)
                    )
                speech = response.json()["response"]["speech"]
                if "plain" in speech:
                    response = speech["plain"]["speech"]
                elif "ssml" in speech:
                    response = speech["ssml"]["speech"]
            if response != "":
                logger.info("Home Assistant response: %s", response)
                syscommand(
                    f'zenity --display="{self.lnxlink.display}" --notification --text="{response}"'
                )
        else:
            data_send = {
                "timevalue": datetime.now().astimezone().isoformat(),
                "attributes": {
                    "key": key,
                },
            }
            self.lnxlink.run_module(self.name, data_send)

    def _test_connection(self):
        try:
            response = get(f"{self.hass_url}/api/", headers=self.hass_headers)
            if response.ok:
                return True
        except Exception:
            return False
        return False



"""Uses xdotool to press keyboard keys"""
import os
import logging
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Send Keys"
        self.lnxlink = lnxlink
        if which("xdotool") is None:
            raise SystemError("System command 'xdotool' not found")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Send Keys": {
                "type": "text",
                "icon": "mdi:keyboard-outline",
            }
        }

    def start_control(self, topic, data):
        """Control system"""
        if os.environ.get("DISPLAY") is None:
            if self.lnxlink.display is not None:
                os.environ["DISPLAY"] = self.lnxlink.display
                logger.info("Initializing empty DISPLAY environment variable")
        syscommand(f"xdotool key {data}")



"""Controls systemd services"""
import logging
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "SystemD"
        self.lnxlink = lnxlink
        if which("systemctl") is None:
            raise SystemError("System command 'systemctl' not found")
        self.services = self.lnxlink.config["settings"].get("systemd", [])
        self.services = [] if self.services is None else self.services
        if len(self.services) == 0:
            logger.info("No systemd settings found on configuration.")

    def get_info(self):
        """Gather information from the system"""
        info = {}
        for service in self.services:
            stdout, _, _ = syscommand(
                f"systemctl show {service} --no-pager | grep ActiveState"
            )
            status = "OFF"
            if "=active" in stdout:
                status = "ON"
            name = service.replace(".service", "")
            info[name] = status
        return info

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for service in self.services:
            name = service.replace(".service", "")
            discovery_info[f"Systemd {name}"] = {
                "type": "switch",
                "icon": "mdi:application-cog",
                "value_template": f"{{{{ value_json.get('{name}') }}}}",
            }
        return discovery_info

    def start_control(self, topic, data):
        """Control system"""
        service = topic[1].replace("systemd_", "")
        if data.lower() == "off":
            syscommand(f"sudo systemctl stop {service}.service &")
        elif data.lower() == "on":
            syscommand(f"sudo systemctl start {service}.service &")



#!/usr/bin/env python3
"""Setup the configuration file"""

import os
import subprocess
import logging
import shutil
from pathlib import Path

import yaml
from lnxlink.consts import SERVICEHEADLESS, SERVICEUSER, CONFIGTEMP

logger = logging.getLogger("lnxlink")


def setup_config(config_path):
    """Setup and create config file"""
    if not os.path.exists(config_path):
        logger.info("Config file not found.")

        try:
            Path(config_path).parent.mkdir(parents=True, exist_ok=True)
            with open(config_path, "wb") as config:
                config.write(CONFIGTEMP.encode())
            logger.info("Created new template: %s", config_path)
        except IOError:
            logger.info("Permision denied")
            return False
        userprompt_config(config_path)
    validate_config(config_path)
    return True


def check_missing(sys_conf, user_conf, missing, dirpath):
    """Recursive method that returns a list of missing dictionary keys"""
    if isinstance(sys_conf, dict):
        for key, value in sys_conf.items():
            check_path = dirpath + [key]
            if isinstance(user_conf, dict) and key in user_conf:
                check_missing(value, user_conf[key], missing, check_path)
            else:
                missing.append([check_path, value])
    return missing


def add_nested(dct, keys, value):
    """
    Adds a nested dictionary item based on a list of keys to an existing dictionary.

    Args:
    dct (dict): The original dictionary to be modified.
    keys (list): A list of keys representing the nested structure.
    value (any): The value to be set at the innermost level.

    Returns:
    dict: The modified dictionary with the new nested item added.
    """
    current_level = dct
    for key in keys[:-1]:
        # Create a new dictionary at the current key if it does not exist
        if key not in current_level:
            current_level[key] = {}
        current_level = current_level[key]

    # Set the value at the innermost level
    if keys[-1] not in current_level:
        current_level[keys[-1]] = value
    return dct


def validate_config(config_path):
    """Inform user of missing configuration values"""
    with open(config_path, "r", encoding="utf8") as file:
        user_conf = yaml.load(file, Loader=yaml.FullLoader)
    sys_conf = yaml.safe_load(CONFIGTEMP)

    missing_keys = check_missing(sys_conf, user_conf, [], [])
    for keys, value in missing_keys:
        key_path = ".".join(keys)
        user_conf = add_nested(user_conf, keys, value)
        logger.info("Added missing configuration option: %s", key_path)

    if len(missing_keys) > 0:
        with open(config_path, "w", encoding="UTF-8") as file:
            file.write(yaml.dump(user_conf, default_flow_style=False, sort_keys=False))


def query_true_false(question, default="false"):
    """Force user to answer yes or no questions"""
    valid = {
        "true": True,
        "t": True,
        "yes": True,
        True: True,
        "false": False,
        "f": False,
        "no": False,
        False: False,
    }

    if default is True:
        prompt = "(True/False) [True]"
    elif default is False or default is None:
        prompt = "(True/False) [False]"
    else:
        raise ValueError(f"invalid default answer: {default}")

    while True:
        choice = input(f" {question} {prompt}: ").lower()
        if default is not None and choice == "":
            return valid[default]
        if choice in valid:
            return valid[choice]
        logger.info("Please respond with 'true' or 'false' (or 't' or 'f').")


def userprompt_config(config_path):
    """Ask users questions to setup config file"""
    with open(config_path, encoding="UTF-8") as file:
        config = yaml.safe_load(file)

        # logger.info config
        logger.info("\nLeave empty for default")

        # Change default values
        config["mqtt"]["discovery"]["enabled"] = query_true_false(
            "Enable MQTT automatic discovery", config["mqtt"]["discovery"]["enabled"]
        )
        config["mqtt"]["server"] = (
            input(f" MQTT server [{config['mqtt']['server']}]: ")
            or config["mqtt"]["server"]
        )
        config["mqtt"]["port"] = (
            input(f" MQTT port [{config['mqtt']['port']}]: ") or config["mqtt"]["port"]
        )
        config["mqtt"]["port"] = int(config["mqtt"]["port"])
        config["mqtt"]["auth"]["tls"] = query_true_false(
            "Enable TLS", config["mqtt"]["auth"]["tls"]
        )
        config["mqtt"]["auth"]["user"] = (
            input(f" MQTT username [{config['mqtt']['auth']['user']}]: ")
            or config["mqtt"]["auth"]["user"]
        )
        config["mqtt"]["auth"]["pass"] = (
            input(f" MQTT password [{config['mqtt']['auth']['pass']}]: ")
            or config["mqtt"]["auth"]["pass"]
        )
        config["mqtt"]["clientId"] = (
            input(f" Change clientId [{config['mqtt']['clientId']}]: ")
            or config["mqtt"]["clientId"]
        )
        statistics = query_true_false("Send statistics", True)
        if not statistics:
            config["exclude"].append("statistics")

    with open(config_path, "w", encoding="UTF-8") as file:
        file.write(yaml.dump(config, default_flow_style=False, sort_keys=False))

    logger.info("\nAll changes have been saved.")
    logger.info(
        " MQTT Topic prefix for for monitoring: %s/%s/...",
        config["mqtt"]["prefix"],
        config["mqtt"]["clientId"],
    )
    logger.info(
        " MQTT Topic prefix for for commands: %s/%s/commands/...",
        config["mqtt"]["prefix"],
        config["mqtt"]["clientId"],
    )


def get_service_user():
    """Install as a user service"""
    installed_as = 0
    for num, cmd_user in enumerate(["--user", ""], start=1):
        cmd = f"systemctl {cmd_user} is-enabled lnxlink.service"
        stdout = subprocess.run(
            cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=False
        ).stdout.decode("UTF-8")
        result = stdout.strip()
        if result in ["enabled"]:
            _, _, _, service_location = get_service_vars(num)
            if os.path.exists(f"{service_location}/lnxlink.service"):
                installed_as = num
    return installed_as


def get_service_vars(user_service):
    """Return service commands based on installation type"""
    if user_service is True or user_service == 1:
        sudo = ""
        cmd_user = "--user"
        systemd_service = SERVICEUSER
        service_location = f"{os.path.expanduser('~')}/.config/systemd/user"
    else:
        sudo = "sudo"
        cmd_user = ""
        systemd_service = SERVICEHEADLESS
        service_location = "/etc/systemd/system"
    return sudo, cmd_user, systemd_service, service_location


def setup_systemd(config_path):
    """Install as a system service"""
    # Check how systemd is installed
    installed_as = get_service_user()

    if installed_as == 0:
        # Service not found or not running
        logger.info("SystemD service not found or it's not running...")
        user_service = query_true_false("Install as a user service?", True)
        sudo, cmd_user, systemd_service, service_location = get_service_vars(
            user_service
        )

        # Install on SystemD
        Path(service_location).mkdir(parents=True, exist_ok=True)
        exec_cmd = f"{shutil.which('lnxlink')} -c {config_path}"
        with open(f"{service_location}/lnxlink.service", "wb") as config:
            config.write(systemd_service.format(exec_cmd=exec_cmd).encode())

        cmd = f"{sudo} chmod +x {service_location}/lnxlink.service"
        subprocess.call(cmd, shell=True)
        cmd = f"{sudo} systemctl {cmd_user} enable lnxlink.service"
        subprocess.call(cmd, shell=True)
        cmd = f"{sudo} systemctl {cmd_user} daemon-reload"
        subprocess.call(cmd, shell=True)


if __name__ == "__main__":
    # config_path = sys.argv[1]
    # download_template(config_path)
    setup_systemd("lnxlink.yaml")



"""Control and show information of currently playing media"""
import re
import hashlib
import traceback
import logging
import base64
from lnxlink.modules.scripts.helpers import import_install_package, syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Media Info"
        self.lnxlink = lnxlink
        self.players = []
        self._requirements()
        self.media_player = self.lib["dbus-mediaplayer"].DBusMediaPlayers(
            self.media_callback
        )

    def _requirements(self):
        self.lib = {
            "alsaaudio": import_install_package("pyalsaaudio", ">=0.9.2", "alsaaudio"),
            "dbus-mediaplayer": import_install_package(
                "dbus-mediaplayer", ">=2024.7.3", "dbus_mediaplayer"
            ),
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Media Info": {
                "type": "sensor",
                "icon": "mdi:music",
                "value_template": "{{ value_json.status }}",
                "attributes_template": "{{ value_json | tojson }}",
            },
            "PlayPause": {
                "type": "button",
                "icon": "mdi:play-pause",
                "enabled": False,
            },
            "Previous": {
                "type": "button",
                "icon": "mdi:skip-previous",
                "enabled": False,
            },
            "Next": {
                "type": "button",
                "icon": "mdi:skip-next",
                "enabled": False,
            },
            "Volume Set": {
                "type": "number",
                "icon": "mdi:volume-high",
                "min": 0,
                "max": 100,
                "enabled": False,
                "value_template": "{{ value_json.volume }}",
            },
            "Thumbnail": {
                "type": "image",
                "encoding": "b64",
                "enabled": False,
                "subtopic": True,
            },
        }

    def start_control(self, topic, data):
        """Control system"""
        if topic[1] == "volume_set":
            mixer = self.lib["alsaaudio"].Mixer()
            if data <= 1:
                data *= 100
            data = min(data, 100)
            mixer.setvolume(int(data))
        elif topic[1] == "playpause":
            if len(self.players) > 0:
                self.media_player.control_media("PlayPause")
        elif topic[1] == "previous":
            if len(self.players) > 0:
                self.media_player.control_media("Previous")
        elif topic[1] == "next":
            if len(self.players) > 0:
                self.media_player.control_media("Next")
        elif topic[1] == "play_media":
            url = data["media_id"]
            syscommand(f"cvlc --play-and-exit {url}", background=True)

    def get_info(self) -> dict:
        """Gather information from the system"""
        info = {
            "title": "",
            "artist": "",
            "album": "",
            "status": "idle",
            "volume": self._get_volume(),
            "playing": False,
            "position": None,
            "duration": None,
        }
        if len(self.players) > 0:
            player = self.players[0]
            info["playing"] = True
            info["title"] = self._filter_title(player["title"])
            info["album"] = player["album"]
            info["artist"] = player["artist"]
            info["status"] = player["status"].lower()
            info["position"] = player["position"]
            info["duration"] = player["duration"]

        return info

    def media_callback(self, players):
        """Callback function to update media information"""
        self.players = players
        self.lnxlink.run_module(f"{self.name}/Thumbnail", self.get_thumbnail)
        self.lnxlink.run_module(self.name, self.get_info)

    def get_thumbnail(self):
        """Returns the thumbnail if it exists as a base64 string"""
        if len(self.players) > 0:
            player = self.players[0]
            try:
                arturl = player["arturl"].replace("file://", "")
                with open(arturl, "rb") as image_file:
                    image_thumbnail = base64.b64encode(image_file.read())
                    return image_thumbnail
            except Exception as err:
                logger.debug(
                    "Can't create thumbnail: %s, %s", err, traceback.format_exc()
                )
        return b" "

    def _get_volume(self):
        """Get system volume"""
        volume = 1
        try:
            mixer = self.lib["alsaaudio"].Mixer()
            volume = mixer.getvolume()[0]
            if mixer.getmute()[0] == 1:
                volume = 0
        except Exception as err:
            logger.error("Can't get volume: %s, %s", err, traceback.format_exc())
        return volume

    def _filter_title(self, title):
        """Returns Title if it contains specific words"""
        if title is None:
            return title
        filter_md5 = [
            "89e55d4f580dd044088b9a003110b37a",
            "7f55a0ed8b021080de00960cc73768fb",
            "912897db01678cc0d2147154dddba25b",
            "2dfc97e5e093b4b339ad510433d2e9ff",
            "0fa9d94de8921144b76e47749b4a168b",
            "c2fc2f64438b1eb36b7e244bdb7bd535",
            "2e1056defdb4eac51954b37a93608628",
            "139ae12f7d434595aa7822bd16b031e2",
            "6e11873b9d9d94a44058bef5747735ce",
            "3c88b51b66c95449d21f55eb9425deff",
            "efde81f569ccb7211e56a522b8b55e5b",
            "1dea8427d5f6f77f13a6b4019b8ae6c1",
            "acc6f2779b808637d04c71e3d8360eeb",
            "347341c085ab1ff63e8409ac9073da9f",
            "cf9a879316551295f08c3b7a94d45598",
            "3c3662bcb661d6de679c636744c66b62",
            "5928db5ba6fb5071370c3ffc22a518a6",
            "348a448a51d1e0f0f5eee42337d12adc",
            "823299e0dbcad6c1e15131c322905248",
            "99754106633f94d350db34d548d6091a",
            "788da20fc6ffb6e232511322ee663555",
            "509ee77f6b418321a66bfd21b64eaac4",
            "fe325cf304ee9155d513be1044bf064b",
            "c34fd2bbae2c55fbb73a23ea25e4769d",
            "ee2a23af409b352d8f1819405bc770b2",
            "d0a87b4018d6521b0f5a49e5e03adb6d",
            "a11c1e7006223a2a80bd295e96566d6e",
            "98a72e2b19d0b3cc63c92a492ded60b3",
            "fd75d7be0587677990a94b446c72b97f",
            "2047a58fb1c23f36d9b74d689bcc495b",
            "c34c8b2c24e2aa2e45d62f1e15967bd2",
            "738615305e5f421c74d2d25b5e9b6620",
            "3221031e8ed5a17ec9a57271de7fadfe",
            "5ffedc3384099eba61237f0a6ee6c9b5",
            "ad0e9d5ed077db5266ff315985114a4f",
            "102ec827703eeb509f12a2a0eb5b4e2d",
            "b6325a80293eb8e2c510a7a8e8468bc1",
            "1841804c061e6f1890f1c7703f2d1bb3",
            "9268d0b2d17670598c70045b0c7abf38",
            "d12761d2356ddb5659c1bad9d46ebec0",
            "36c029068925a76bbaa77670d6754841",
            "f7c35a9cfd54951862e895bbebef145f",
            "2b6d39e4dc4cdf1ebe3f1a2e80f18202",
            "7bc66d4625c71a3bd3dd6e1505050616",
            "6a1fccefbc7f074b5017884743cde766",
            "fe81a4f28e6bd176efc8184d58544e66",
        ]
        words = re.split(r";|,|\*|\n| +|you|\.|\/|ing", title.strip().lower())
        for word in words:
            md5_word = hashlib.md5(word.encode("UTF-8")).hexdigest()
            if md5_word in filter_md5:
                return "Title"
        return title



"""Checks if restart is needed"""
import os


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Required Restart"

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Required Restart": {
                "type": "binary_sensor",
                "icon": "mdi:alert-octagon-outline",
                "entity_category": "diagnostic",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        if os.path.exists("/var/run/reboot-required"):
            return "ON"
        return "OFF"



"""Controls the brightness of the displays"""
import re
import logging
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink=None):
        """Setup addon"""
        self.name = "Brightness"
        self.lnxlink = lnxlink
        if which("xrandr") is None:
            raise SystemError("System command 'xrandr' not found")
        self.displays = self._get_displays()

    def get_info(self):
        """Gather information from the system"""
        displays = self._get_displays()
        if displays != self.displays:
            self.displays = displays
            self.lnxlink.setup_discovery()
        values = [disp["brightness"] for disp in displays.values()]
        avg_brightness = sum(values) / max(1, len(values))
        avg_brightness = max(0.1, avg_brightness)

        info = {"status": avg_brightness}
        for display, values in displays.items():
            info[display] = max(0.1, values["brightness"])
        return info

    def exposed_controls(self):
        """Exposes to home assistant"""
        controls = {}
        if len(self.displays) > 0:
            controls = {
                "Brightness": {
                    "type": "number",
                    "icon": "mdi:brightness-7",
                    "min": 0.1,
                    "max": 1,
                    "step": 0.1,
                    "value_template": "{{ value_json.status }}",
                }
            }
        for display, values in self.displays.items():
            controls[f"Brightness {values['name']}"] = {
                "type": "number",
                "icon": "mdi:brightness-7",
                "min": 0.1,
                "max": 1,
                "step": 0.1,
                "value_template": f"{{{{ value_json.{display} }}}}",
                "enabled": False,
            }
        return controls

    def start_control(self, topic, data):
        """Control system"""
        disp_env_cmd = ""
        if self.lnxlink.display is not None:
            disp_env_cmd = f" --display {self.lnxlink.display}"

        if topic[1] == "brightness":
            for values in self.displays.values():
                syscommand(
                    f"xrandr --output {values['name']} --brightness {data} {disp_env_cmd}"
                )
        else:
            display = self.displays[
                topic[1].replace("brightness_", "").replace("-", "_")
            ]["name"]
            syscommand(f"xrandr --output {display} --brightness {data} {disp_env_cmd}")

    def _get_displays(self):
        """Get all the displays"""
        displays = {}
        disp_env_cmd = ""
        if self.lnxlink.display is not None:
            disp_env_cmd = f" --display {self.lnxlink.display}"

        stdout, _, _ = syscommand(
            f"xrandr --verbose --current {disp_env_cmd}",
        )
        pattern = re.compile(
            r"(\S+) \bconnected\b.*[\s\S]*?(?=Brightness)Brightness: ([\d\.\d]+)"
        )

        for match in pattern.findall(stdout):
            displays[match[0].replace("-", "_").lower()] = {
                "name": match[0],
                "brightness": float(match[1]),
            }

        return displays



"""Sends statistics information"""
import os
import uuid
import json
import logging
import time
import requests

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Statistics"
        self.lnxlink = lnxlink
        # Check for updates every 24 hours
        self.update_interval = 86400
        # 15 minutes after lnxlink starts
        self.last_time = time.time() - self.update_interval + 900
        self.url = self.lnxlink.config["settings"]["statistics"]
        if self.url is None or self.url == "":
            raise SystemError("Statistics is not setup correctly")
        self.uuid = self._get_uuid()

    def get_info(self):
        """Gather information from the system"""
        cur_time = time.time()
        if cur_time - self.last_time > self.update_interval:
            version = self.lnxlink.version.split("+")[0]
            data = json.dumps(
                {
                    "uuid": self.uuid,
                    "version": version,
                }
            )
            logger.info("Sending statistics data: %s", data)
            url = self.url.rstrip("/")
            url = f"{url}/api/lnxlink"
            requests.post(url=url, data=data)
            self.last_time = cur_time

    def _get_uuid(self):
        """Creates a Unique ID once to send it to the analyzer server"""
        config_dir = os.path.dirname(os.path.realpath(self.lnxlink.config_path))
        uuid_file = os.path.join(config_dir, "lnxlink_uuid.txt")
        if os.path.exists(uuid_file):
            with open(uuid_file, encoding="UTF-8") as file:
                return file.read()
        else:
            tmp_uuid = uuid.uuid4().hex
            with open(uuid_file, "w", encoding="UTF-8") as file:
                file.write(tmp_uuid)
            return tmp_uuid



"""Gets temperatures"""
import logging

import psutil

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Temperature"
        self.lnxlink = lnxlink
        self.temperatures = self._get_temperatures()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for key, value in self.temperatures.items():
            discovery_info[f"Temperature {value['name']}"] = {
                "type": "sensor",
                "entity_category": "diagnostic",
                "state_class": "measurement",
                "device_class": "temperature",
                "unit": "°C",
                "value_template": f"{{{{ value_json.get('{key}') }}}}",
                "enabled": True,
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        temperatures = self._get_temperatures()
        temperatures = {key: value["value"] for key, value in temperatures.items()}
        return temperatures

    def _get_temperatures(self):
        """Get a list of all temperatures"""
        temperatures = {}
        for item, values in psutil.sensors_temperatures().items():
            for value in values:
                name = f"{item} {value.label}"
                key = name.replace(" ", "_").lower()
                temperatures[key] = {
                    "name": name,
                    "value": value.current,
                }
        return temperatures
